import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.pad import pad_array as pad_array_function
from gdsfactory.port import select_ports_electrical
from gdsfactory.routing.route_quad import route_quad
from gdsfactory.types import ComponentFactory, Float2


@gf.cell
def add_electrical_pads_top(
    component: Component,
    spacing: Float2 = (0.0, 100.0),
    pad_array: ComponentFactory = pad_array_function,
    select_ports=select_ports_electrical,
    layer: gf.types.Layer = (31, 0),
) -> Component:
    """Returns new component with electrical ports connected to top pad array

    Args:
        component:
        spacing: component to pad spacing
        pad_array: function for pad_array
        select_ports: function to select electrical ports
        layer: for the routes
    """
    c = Component()
    c.component = component
    ref = c << component
    ports = select_ports(ref.ports)
    ports = list(ports.values())
    pads = c << pad_array(columns=len(ports), orientation=270)
    pads.x = ref.x + spacing[0]
    pads.ymin = ref.ymax + spacing[1]
    ports_pads = list(pads.ports.values())

    ports_pads = gf.routing.sort_ports.sort_ports_x(ports_pads)
    ports_component = gf.routing.sort_ports.sort_ports_x(ports)

    for p1, p2 in zip(ports_component, ports_pads):
        c.add(route_quad(p1, p2, layer=layer))

    c.add_ports(ref.ports)
    for port in ports:
        c.ports.pop(port.name)
    c.copy_child_info(component)
    return c


if __name__ == "__main__":
    # FIXME
    # c = demo_mzi()
    # c = demo_straight()
    # c.show()
    import gdsfactory as gf

    c = gf.components.straight_heater_metal()
    c = gf.components.mzi_phase_shifter()
    cc = add_electrical_pads_top(component=c, spacing=(-150, 30))
    cc.show()



from typing import Optional

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.contact import contact_heater_m3
from gdsfactory.components.coupler_ring import coupler_ring as coupler_ring_function
from gdsfactory.components.straight import straight as straight_function
from gdsfactory.config import call_if_func
from gdsfactory.cross_section import strip
from gdsfactory.snap import assert_on_2nm_grid
from gdsfactory.types import ComponentFactory, CrossSectionFactory, Float2

contact_heater_m3_mini = gf.partial(contact_heater_m3, size=(4, 4))


@gf.cell
def ring_double_heater(
    gap: float = 0.2,
    radius: float = 10.0,
    length_x: float = 0.01,
    length_y: float = 0.01,
    coupler_ring: ComponentFactory = coupler_ring_function,
    straight: ComponentFactory = straight_function,
    bend: Optional[ComponentFactory] = None,
    cross_section_heater: gf.types.CrossSectionFactory = gf.cross_section.strip_heater_metal,
    cross_section: CrossSectionFactory = strip,
    contact: gf.types.ComponentFactory = contact_heater_m3_mini,
    port_orientation: int = 90,
    contact_offset: Float2 = (0, 0),
    **kwargs
) -> Component:
    """Double bus ring made of two couplers (ct: top, cb: bottom)
    connected with two vertical straights (sl: left, sr: right)
    includes heater on top

    Args:
        gap: gap between for coupler
        radius: for the bend and coupler
        length_x: ring coupler length
        length_y: vertical straight length
        coupler_ring: ring coupler function
        straight: straight function
        bend: bend function
        cross_section_heater:
        cross_section:
        contact:
        port_orientation: for electrical ports to promote from contact
        contact_offset: for each contact
        kwargs: cross_section settings

    .. code::

         --==ct==--
          |      |
          sl     sr length_y
          |      |
         --==cb==-- gap

          length_x

    """
    assert_on_2nm_grid(gap)

    coupler_component = (
        coupler_ring(
            gap=gap,
            radius=radius,
            length_x=length_x,
            bend=bend,
            cross_section=cross_section,
            bend_cross_section=cross_section_heater,
            **kwargs
        )
        if callable(coupler_ring)
        else coupler_ring
    )
    straight_component = call_if_func(
        straight, length=length_y, cross_section=cross_section_heater, **kwargs
    )

    c = Component()
    cb = c.add_ref(coupler_component)
    ct = c.add_ref(coupler_component)
    sl = c.add_ref(straight_component)
    sr = c.add_ref(straight_component)

    sl.connect(port="o1", destination=cb.ports["o2"])
    ct.connect(port="o3", destination=sl.ports["o2"])
    sr.connect(port="o2", destination=ct.ports["o2"])
    c.add_port("o1", port=cb.ports["o1"])
    c.add_port("o2", port=cb.ports["o4"])
    c.add_port("o3", port=ct.ports["o4"])
    c.add_port("o4", port=ct.ports["o1"])

    c1 = c << contact()
    c2 = c << contact()
    c1.xmax = -length_x / 2 + cb.x - contact_offset[0]
    c2.xmin = +length_x / 2 + cb.x + contact_offset[0]
    c1.movey(contact_offset[1])
    c2.movey(contact_offset[1])
    c.add_ports(c1.get_ports_list(orientation=port_orientation), prefix="e1")
    c.add_ports(c2.get_ports_list(orientation=port_orientation), prefix="e2")
    c.auto_rename_ports()
    return c


if __name__ == "__main__":

    c = ring_double_heater(width=1, layer=(2, 0), length_y=3)
    c.show(show_subports=False)



"""Sample AWG."""
import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.cross_section import strip
from gdsfactory.types import CrossSectionFactory


@gf.cell
def free_propagation_region(
    width1: float = 2.0,
    width2: float = 20.0,
    length: float = 20.0,
    wg_width: float = 0.5,
    inputs: int = 1,
    outputs: int = 10,
    cross_section: CrossSectionFactory = strip,
    **kwargs,
) -> Component:
    r"""

    .. code::

                 length
                 <-->
                   /|
                  / |
           width1|  | width2
                  \ |
                   \|
    """
    y1 = width1 / 2
    y2 = width2 / 2
    x = cross_section(**kwargs)
    o = x.info["cladding_offset"]
    layers_cladding = x.info["layers_cladding"] or []
    layer = x.info["layer"]

    xpts = [0, length, length, 0]
    ypts = [y1, y2, -y2, -y1]

    c = gf.Component()
    c.add_polygon((xpts, ypts), layer=layer)

    if inputs == 1:
        c.add_port(
            "o1",
            midpoint=(0, 0),
            width=wg_width,
            orientation=180,
            layer=layer,
        )
    else:
        y = np.linspace(-width1 / 2 + wg_width / 2, width1 / 2 - wg_width / 2, inputs)
        y = gf.snap.snap_to_grid(y)
        for i, y in enumerate(y):
            c.add_port(
                f"W{i}",
                midpoint=(0, y),
                width=wg_width,
                orientation=0,
                layer=layer,
            )

    y = np.linspace(-width2 / 2 + wg_width / 2, width2 / 2 - wg_width / 2, outputs)
    y = gf.snap.snap_to_grid(y)
    for i, y in enumerate(y):
        c.add_port(
            f"E{i}",
            midpoint=(length, y),
            width=wg_width,
            orientation=0,
            layer=layer,
        )

    ypts = [y1 + o, y2 + o, -y2 - o, -y1 - o]

    for layer in layers_cladding:
        c.add_polygon((xpts, ypts), layer=layer)

    c.info["length"] = length
    c.info["width1"] = width1
    c.info["width2"] = width2
    return c


@gf.cell
def free_propagation_region_input(inputs: int = 1, **kwargs) -> Component:
    return free_propagation_region(inputs=inputs, **kwargs)


@gf.cell
def free_propagation_region_output(
    inputs: int = 10, width1: float = 10.0, width2: float = 20.0, **kwargs
) -> Component:
    return free_propagation_region(
        inputs=inputs, width2=width2, width1=width1, **kwargs
    )


@gf.cell
def awg(
    arms: int = 10,
    outputs: int = 3,
    free_propagation_region_input_function=free_propagation_region_input,
    free_propagation_region_output_function=free_propagation_region_output,
    fpr_spacing: float = 50.0,
) -> Component:
    """Returns a basic Arrayed Waveguide grating.

    Args:
        arms: number of arms
        outputs: number of outputs
        free_propagation_region_input_function: for input
        free_propagation_region_output_function: for output
        fpr_spacing: x separation between input/output FPR

    """
    c = Component()
    fpr_in = free_propagation_region_input_function(
        inputs=1,
        outputs=arms,
    )
    fpr_out = free_propagation_region_output_function(
        inputs=outputs,
        outputs=arms,
    )

    fpr_in_ref = c.add_ref(fpr_in)
    fpr_out_ref = c.add_ref(fpr_out)

    fpr_in_ref.rotate(90)
    fpr_out_ref.rotate(90)

    fpr_out_ref.x += fpr_spacing
    routes = gf.routing.get_bundle(
        fpr_in_ref.get_ports_list(prefix="E"), fpr_out_ref.get_ports_list(prefix="E")
    )

    c.lengths = []
    for route in routes:
        c.add(route.references)
        c.lengths.append(route.length)

    c.add_port("o1", port=fpr_in_ref.ports["o1"])

    for i, port in enumerate(fpr_out_ref.get_ports_list(prefix="W")):
        c.add_port(f"E{i}", port=port)

    c.delta_length = np.mean(np.diff(c.lengths))
    return c


if __name__ == "__main__":
    c = free_propagation_region(inputs=2, outputs=4)
    # print(c.ports.keys())
    c = awg()
    c.show()



"""You add pins in a pin layer to clearly see the component ports."""

import gdsfactory as gf


@gf.cell
def straight_sample(length=5, width=1):
    """Returns straight with ports."""
    wg = gf.Component("straight_sample")
    wg.add_polygon([(0, 0), (length, 0), (length, width), (0, width)], layer=(1, 0))
    wg.add_port(name="o1", midpoint=[0, width / 2], width=width, orientation=180)
    wg.add_port(name="o2", midpoint=[length, width / 2], width=width, orientation=0)
    return wg


if __name__ == "__main__":
    wg = straight_sample()
    gf.add_pins(wg)
    wg.show()



from typing import List, Optional

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.port import Port
from gdsfactory.types import ComponentOrFactory, Strs


@cell
def extend_ports_list(
    ports: List[Port],
    extension_factory: ComponentOrFactory,
    extension_port_name: Optional[str] = None,
    ignore_ports: Optional[Strs] = None,
) -> Component:
    """Returns a component with the extensions for a list of ports.

    Args:
        ports: list of ports
        extension_factory: function for extension
        extension_port_name: to connect extension
        ignore_ports: list of port names to ignore

    """
    c = Component()
    extension = (
        extension_factory() if callable(extension_factory) else extension_factory
    )

    extension_port_name = extension_port_name or list(extension.ports.keys())[0]
    ignore_ports = ignore_ports or []

    for i, port in enumerate(ports):
        extension_ref = c << extension
        extension_ref.connect(extension_port_name, port)

        for port_name, port in extension_ref.ports.items():
            if port_name not in ignore_ports:
                c.add_port(f"{i}_{port_name}", port=port)

    c.auto_rename_ports()
    return c


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.components.mmi1x2()
    t = gf.partial(gf.components.taper, width2=0.1)

    cr = extend_ports_list(
        ports=c.get_ports_list(), extension_factory=t, extension_port_name="o1"
    )
    c.add_ref(cr)
    c.show()



from gdsfactory.mask.merge_json import merge_json
from gdsfactory.mask.merge_markdown import merge_markdown
from gdsfactory.mask.merge_metadata import merge_metadata
from gdsfactory.mask.merge_test_metadata import get_cell_from_label, merge_test_metadata
from gdsfactory.mask.merge_yaml import merge_yaml
from gdsfactory.mask.write_labels import find_labels, write_labels

__all__ = [
    "find_labels",
    "get_cell_from_label",
    "merge_json",
    "merge_yaml",
    "merge_markdown",
    "merge_metadata",
    "merge_test_metadata",
    "write_labels",
]



import pathlib
import tempfile
from typing import Optional, Tuple, Union

import meep as mp
import numpy as np
import pydantic
from meep import mpb

mpb.Verbosity(0)

tmp = pathlib.Path(tempfile.TemporaryDirectory().name).parent / "meep"
tmp.mkdir(exist_ok=True)

Floats = Tuple[float, ...]


@pydantic.validate_arguments
def get_mode_solver_coupler(
    wg_width: float = 0.5,
    gap: float = 0.2,
    wg_widths: Optional[Floats] = None,
    gaps: Optional[Floats] = None,
    wg_thickness: float = 0.22,
    slab_thickness: float = 0.0,
    ncore: float = 3.47,
    nclad: float = 1.44,
    nslab: Optional[float] = None,
    ymargin: float = 2.0,
    sz: float = 2.0,
    resolution: int = 32,
    nmodes: int = 4,
    sidewall_angles: Union[Tuple[float, ...], float] = None,
    # sidewall_taper: int = 1,
) -> mpb.ModeSolver:
    """Returns a mode_solver simulation.

    Args:
        wg_width: wg_width (um)
        gap:
        wg_widths: list or tuple of waveguide widths.
        gaps: list or tuple of waveguide gaps.
        wg_thickness: wg height (um)
        slab_thickness: thickness for the waveguide slab
        ncore: core material refractive index
        nclad: clad material refractive index
        nslab: Optional slab material refractive index. Defaults to ncore.
        ymargin: margin in y.
        sz: simulation region thickness (um)
        resolution: resolution (pixels/um)
        nmodes: number of modes
        sidewall_angles: waveguide sidewall angle (radians),
            tapers from wg_width at top of slab, upwards, to top of waveguide

    ::

          _____________________________________________________
          |
          |
          |         widths[0]                 widths[1]
          |     <---------->     gaps[0]    <---------->
          |      ___________ <-------------> ___________      _
          |     |           |               |           |     |
        sz|_____|  ncore    |_______________|           |_____|
          |                                                   | wg_thickness
          |slab_thickness        nslab                        |
          |___________________________________________________|
          |
          |<--->                                         <--->
          |ymargin               nclad                   ymargin
          |____________________________________________________
          <--------------------------------------------------->
                                   sy



    """
    wg_widths = wg_widths or (wg_width, wg_width)
    gaps = gaps or (gap,)
    material_core = mp.Medium(index=ncore)
    material_clad = mp.Medium(index=nclad)
    material_slab = mp.Medium(index=nslab or ncore)

    # Define the computational cell.  We'll make x the propagation direction.
    # the other cell sizes should be big enough so that the boundaries are
    # far away from the mode field.

    sy = np.sum(wg_widths) + np.sum(gaps) + 2 * ymargin
    geometry_lattice = mp.Lattice(size=mp.Vector3(0, sy, sz))

    geometry = []

    y = -sy / 2 + ymargin

    gaps = list(gaps) + [0]
    for i, wg_width in enumerate(wg_widths):
        if sidewall_angles:
            geometry.append(
                mp.Prism(
                    vertices=[
                        mp.Vector3(y=y, z=slab_thickness),
                        mp.Vector3(y=y + wg_width, z=slab_thickness),
                        mp.Vector3(x=1, y=y + wg_width, z=slab_thickness),
                        mp.Vector3(x=1, y=y, z=slab_thickness),
                    ],
                    height=wg_thickness - slab_thickness,
                    center=mp.Vector3(
                        y=y + wg_width / 2,
                        z=slab_thickness + (wg_thickness - slab_thickness) / 2,
                    ),
                    # If only 1 angle is specified, use it for all waveguides
                    sidewall_angle=sidewall_angles
                    if len(np.unique(sidewall_angles)) == 1
                    else sidewall_angles[i],
                    # axis=mp.Vector3(z=sidewall_taper),
                    material=material_core,
                )
            )
        else:
            geometry.append(
                mp.Block(
                    size=mp.Vector3(mp.inf, wg_width, wg_thickness),
                    material=material_core,
                    center=mp.Vector3(y=y + wg_width / 2, z=wg_thickness / 2),
                )
            )

        y += gaps[i] + wg_width

    # define the 2D blocks for the strip and substrate
    geometry += [
        mp.Block(
            size=mp.Vector3(mp.inf, mp.inf, slab_thickness),
            material=material_slab,
            center=mp.Vector3(z=slab_thickness / 2),
        ),
    ]

    # The k (i.e. beta, i.e. propagation constant) points to look at, in
    # units of 2*pi/um.  We'll look at num_k points from k_min to k_max.
    num_k = 9
    k_min = 0.1
    k_max = 3.0
    k_points = mp.interpolate(num_k, [mp.Vector3(k_min), mp.Vector3(k_max)])

    # Increase this to see more modes.  (The guided ones are the ones below the
    # light line, i.e. those with frequencies < kmag / 1.45, where kmag
    # is the corresponding column in the output if you grep for "freqs:".)
    # use this prefix for output files

    wg_widths_str = "_".join([str(i) for i in wg_widths])
    gaps_str = "_".join([str(i) for i in gaps])
    filename_prefix = (
        tmp / f"coupler_{wg_widths_str}_{gaps_str}_{wg_thickness}_{slab_thickness}"
    )

    mode_solver = mpb.ModeSolver(
        geometry_lattice=geometry_lattice,
        geometry=geometry,
        k_points=k_points,
        resolution=resolution,
        num_bands=nmodes,
        filename_prefix=str(filename_prefix),
        default_material=material_clad,
    )
    mode_solver.nmodes = nmodes
    mode_solver.info = dict(
        wg_widths=wg_widths,
        gaps=gaps,
        wg_thickness=wg_thickness,
        slab_thickness=slab_thickness,
        ncore=ncore,
        nclad=nclad,
        sy=sy,
        sz=sz,
        resolution=resolution,
        nmodes=nmodes,
    )
    return mode_solver


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    m = get_mode_solver_coupler(
        slab_thickness=90e-3,
        nslab=2,
        gap=0.5,
        wg_width=1,
        resolution=64,
        sidewall_angles=(10.0 * (np.pi / 180), 20.0 * (np.pi / 180)),
    )
    m.init_params(p=mp.NO_PARITY, reset_fields=False)
    eps = m.get_epsilon()
    # cmap = 'viridis'
    # cmap = "RdBu"
    cmap = "binary"
    origin = "lower"
    plt.imshow(
        eps.T ** 0.5,
        cmap=cmap,
        origin=origin,
        aspect="auto",
        extent=[
            -m.info["sy"] / 2,
            m.info["sy"] / 2,
            -m.info["sz"] / 2,
            m.info["sz"] / 2,
        ],
    )
    plt.colorbar()
    plt.show()



"""gdsfactory tidy3d plugin

[tidy3D is a fast GPU based commercial FDTD solver](https://simulation.cloud/)
"""

from gdsfactory.simulation.gtidy3d import materials
from gdsfactory.simulation.gtidy3d.get_results import get_results
from gdsfactory.simulation.gtidy3d.get_simulation import (
    get_simulation,
    plot_simulation,
    plot_simulation_xz,
    plot_simulation_yz,
)
from gdsfactory.simulation.gtidy3d.get_sparameters import get_sparameters

__version__ = "0.0.2"
__all__ = [
    "plot_simulation",
    "plot_simulation_xz",
    "plot_simulation_yz",
    "get_sparameters",
    "get_simulation",
    "get_results",
    "materials",
]



import gdsfactory as gf
from gdsfactory.add_padding import add_padding
from gdsfactory.component import Component
from gdsfactory.components.taper import taper as taper_function
from gdsfactory.cross_section import strip
from gdsfactory.types import ComponentFactory, CrossSectionFactory


@gf.cell
def mmi2x2(
    width: float = 0.5,
    width_taper: float = 1.0,
    length_taper: float = 10.0,
    length_mmi: float = 5.5,
    width_mmi: float = 2.5,
    gap_mmi: float = 0.25,
    taper: ComponentFactory = taper_function,
    cross_section: CrossSectionFactory = strip,
    **kwargs
) -> Component:
    r"""Mmi 2x2.

    Args:
        width: input and output straight width
        width_taper: interface between input straights and mmi region
        length_taper: into the mmi region
        length_mmi: in x direction
        width_mmi: in y direction
        gap_mmi: (width_taper + gap between tapered wg)/2
        taper: taper function
        cross_section:
        **kwargs: cross_section settings


    .. code::

                   length_mmi
                    <------>
                    ________
                   |        |
                __/          \__
            W1  __            __  E1
                  \          /_ _ _ _
                  |         | _ _ _ _| gap_mmi
                __/          \__
            W0  __            __  E0
                  \          /
                   |________|

                 <->
            length_taper

    """
    gf.snap.assert_on_2nm_grid(gap_mmi)
    x = cross_section(**kwargs)
    cladding_offset = x.info["cladding_offset"]
    layers_cladding = x.info["layers_cladding"]
    layer = x.info["layer"]

    component = gf.Component()
    w_mmi = width_mmi
    w_taper = width_taper

    taper = taper(
        length=length_taper,
        width1=width,
        width2=w_taper,
        cross_section=cross_section,
        **kwargs
    )

    a = gap_mmi / 2 + width_taper / 2
    mmi = gf.components.rectangle(
        size=(length_mmi, w_mmi),
        layer=layer,
        centered=True,
    )

    ports = [
        gf.Port("o1", orientation=180, midpoint=(-length_mmi / 2, -a), width=w_taper),
        gf.Port("o2", orientation=180, midpoint=(-length_mmi / 2, +a), width=w_taper),
        gf.Port("o3", orientation=0, midpoint=(+length_mmi / 2, +a), width=w_taper),
        gf.Port("o4", orientation=0, midpoint=(+length_mmi / 2, -a), width=w_taper),
    ]

    mmi_section = component.add_ref(mmi)

    for port in ports:
        taper_ref = component << taper
        taper_ref.connect(port="o2", destination=port)
        component.add_port(name=port.name, port=taper_ref.ports["o1"])
        component.absorb(taper_ref)

    component.absorb(mmi_section)

    layers_cladding = layers_cladding or []
    if layers_cladding:
        add_padding(
            component,
            default=cladding_offset,
            right=0,
            left=0,
            top=cladding_offset,
            bottom=cladding_offset,
            layers=layers_cladding,
        )
    component.auto_rename_ports()
    return component


if __name__ == "__main__":
    c = mmi2x2(layer=(2, 0), gap_mmi=0.251)
    c.show()
    c.pprint()



"""You can define a path with a list of points

To create a component you need to extrude the path with a cross-section.

"""
from functools import partial
from typing import Optional, Tuple

import pydantic
from phidl.device_layout import CrossSection as CrossSectionPhidl

from gdsfactory.tech import TECH, Section

LAYER = TECH.layer
Layer = Tuple[int, int]
Layers = Tuple[Layer, ...]
Floats = Tuple[float, ...]


class CrossSection(CrossSectionPhidl):
    """Extend phidl.device_layout.CrossSection with port_types.

    Args:
        sections: list of sections
        ports:
        port_types:
        aliases:
        info:
        name:

    """

    def __init__(self):
        self.sections = []
        self.ports = set()
        self.port_types = set()
        self.aliases = {}
        self.info = {}
        self.name = None

    def add(
        self,
        width: float = 1,
        offset: float = 0,
        layer: Tuple[int, int] = (1, 0),
        ports: Tuple[Optional[str], Optional[str]] = (None, None),
        name: Optional[str] = None,
        port_types: Tuple[str, str] = ("optical", "optical"),
        hidden: bool = False,
    ):
        """Adds a cross-sectional element to the CrossSection. If ports are
        specified, when creating a Component with the gf.path.extrude() command
        it will add Ports at the ends.

        Args:
            width: width of the segment (um)
            offset: Offset of the segment (positive values = right hand side)
            layer: The polygon layer to put the segment on
            ports: Optional port names at the ends of the cross-section
            name: Optional Name of the cross-sectional element for later access
            port_types: electrical, optical ...
            hidden: if True does not draw polygons for CrossSection

        .. code::

              0   offset
              |<-------------->|
              |              _____
              |             |     |
              |             |layer|
              |             |_____|
              |              <---->
                             width

        """
        if isinstance(width, (float, int)) and (width <= 0):
            raise ValueError("CrossSection.add(): widths must be >0")
        if len(ports) != 2:
            raise ValueError("CrossSection.add(): must receive 2 port names")
        for p in ports:
            if p is not None and p in self.ports:
                raise ValueError(
                    f"CrossSection.add(): a port named {p!r} already "
                    "exists in this CrossSection, please rename port"
                )

        [self.ports.add(port) for port in ports if port is not None]
        [self.port_types.add(port_type) for port_type in port_types]

        if name in self.aliases:
            raise ValueError(
                f"CrossSection.add(): an element named {name!r} already "
                "exists in this CrossSection, please change the name"
            )

        new_segment = dict(
            width=width,
            offset=offset,
            layer=layer,
            ports=ports,
            port_types=port_types,
            hidden=hidden,
            name=name,
        )

        if name is not None:
            self.aliases[name] = new_segment
        self.sections.append(new_segment)
        return self

    def copy(self):
        """Return a copy of the CrossSection"""
        X = CrossSection()
        X.info = self.info.copy()
        X.sections = list(self.sections)
        X.ports = tuple(sorted(list(self.ports)))
        X.aliases = dict(self.aliases)
        X.port_types = tuple(self.port_types)
        return X

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        """pydantic checks CrossSection valid type"""
        assert isinstance(
            v, CrossSection
        ), f"TypeError, Got {type(v)}, expecting CrossSection"
        return v

    def to_dict(self):
        d = {}
        x = self.copy()
        d["sections"] = [dict(section) for section in x.sections if section]
        d["ports"] = x.ports
        d["port_types"] = x.port_types
        d["aliases"] = x.aliases
        d["info"] = x.info
        if hasattr(self, "cross_sections"):
            d["cross_sections"] = [x.to_dict() for x in self.cross_sections]
        return d

    def get_name(self):
        return self.name or "_".join([str(i) for i in self.to_dict()["sections"]])


@pydantic.validate_arguments
def cross_section(
    width: float = 0.5,
    layer: Tuple[int, int] = (1, 0),
    width_wide: Optional[float] = None,
    auto_widen: bool = False,
    auto_widen_minimum_length: float = 200.0,
    taper_length: float = 10.0,
    radius: float = 10.0,
    cladding_offset: float = 3.0,
    layers_cladding: Optional[Tuple[Layer, ...]] = None,
    sections: Optional[Tuple[Section, ...]] = None,
    port_names: Tuple[str, str] = ("o1", "o2"),
    port_types: Tuple[str, str] = ("optical", "optical"),
    min_length: float = 10e-3,
    start_straight_length: float = 10e-3,
    end_straight_length: float = 10e-3,
    snap_to_grid: Optional[float] = None,
) -> CrossSection:
    """Return CrossSection.

    Args:
        width: main of the waveguide
        layer: main layer
        width_wide: taper to widen waveguides for lower loss
        auto_widen: taper to widen waveguides for lower loss
        auto_widen_minimum_length: minimum straight length for auto_widen
        taper_length: taper_length for auto_widen
        radius: bend radius
        cladding_offset: offset for layers_cladding
        layers_cladding:
        sections: Sections(width, offset, layer, ports)
        port_names: for input and output ('o1', 'o2')
        port_types: for input and output (electrical, optical, vertical_te ...)
        min_length: 10e-3 for routing
        start_straight_length: for routing
        end_straight_length: for routing
        snap_to_grid: can snap points to grid when extruding the path
    """

    x = CrossSection()
    x.add(
        width=width,
        offset=0,
        layer=layer,
        ports=port_names,
        port_types=port_types,
        name="_default",
    )

    sections = sections or []
    for section in sections:
        if isinstance(section, dict):
            x.add(
                width=section["width"],
                offset=section["offset"],
                layer=section["layer"],
                ports=section["ports"],
                port_types=section["port_types"],
                name=section["name"],
            )
        else:
            x.add(
                width=section.width,
                offset=section.offset,
                layer=section.layer,
                ports=section.ports,
                port_types=section.port_types,
                name=section.name,
            )

    x.info = dict(
        width=width,
        layer=layer,
        width_wide=width_wide,
        auto_widen=auto_widen,
        auto_widen_minimum_length=auto_widen_minimum_length,
        taper_length=taper_length,
        radius=radius,
        cladding_offset=cladding_offset,
        layers_cladding=layers_cladding,
        sections=sections,
        min_length=min_length,
        start_straight_length=start_straight_length,
        end_straight_length=end_straight_length,
        snap_to_grid=snap_to_grid,
        port_types=port_types,
        port_names=port_names,
    )
    return x


@pydantic.validate_arguments
def pin(
    width: float = 0.5,
    layer: Tuple[int, int] = LAYER.WG,
    layer_slab: Tuple[int, int] = LAYER.SLAB90,
    layers_contact1: Layers = (LAYER.PPP,),
    layers_contact2: Layers = (LAYER.NPP,),
    cladding_offsets_contact1: Tuple[float, ...] = (0, -0.2),
    cladding_offsets_contact2: Tuple[float, ...] = (0, -0.2),
    contact_width: float = 9.0,
    contact_gap: float = 0.55,
    slab_gap: float = -0.2,
    layer_via: Optional[Layer] = None,
    via_width: float = 1,
    via_offsets: Optional[Tuple[float, ...]] = None,
    **kwargs,
) -> CrossSection:
    """rib PIN doped cross_section.

    Args:
        width: ridge width
        layer: ridge layer
        layer_slab: slab layer
        layers_contact1: P++ layer
        layers_contact2: N++ layer
        cladding_offsets_contact1:
        cladding_offsets_contact2:
        contact_width:
        contact_gap: offset from contact to ridge edge
        slab_gap: extra slab gap (negative: contact goes beyond slab)
        layer_via:
        via_width:
        via_offsets:
        kwargs: other cross_section settings

    https://doi.org/10.1364/OE.26.029983

    .. code::

                                         layer
                                 |<------width------>|
                                  ____________________ contact_gap              slab_gap
                                 |                   |<----------->|             <-->
        ___ _____________________|                   |__________________________|___
       |   |         |                 undoped Si                  |            |   |
       |   |    P++  |                 intrinsic region            |     N++    |   |
       |___|_________|_____________________________________________|____________|___|
                                                                    <----------->
                                                                    contact_width
       <---------------------------------------------------------------------------->
                                      slab_width
    """
    slab_width = width + 2 * contact_gap + 2 * contact_width - 2 * slab_gap
    contact_offset = width / 2 + contact_gap + contact_width / 2

    sections = [Section(width=slab_width, layer=layer_slab, name="slab")]
    sections += [
        Section(
            layer=layer,
            width=contact_width + 2 * cladding_offset,
            offset=+contact_offset,
        )
        for layer, cladding_offset in zip(layers_contact1, cladding_offsets_contact1)
    ]
    sections += [
        Section(
            layer=layer,
            width=contact_width + 2 * cladding_offset,
            offset=-contact_offset,
        )
        for layer, cladding_offset in zip(layers_contact2, cladding_offsets_contact2)
    ]

    if layer_via and via_width and via_offsets:
        sections += [
            Section(
                layer=layer_via,
                width=via_width,
                offset=offset,
            )
            for offset in via_offsets
        ]

    x = cross_section(
        width=width,
        layer=layer,
        sections=sections,
        **kwargs,
    )
    info = dict(
        width=width,
        layer=layer,
        layer_slab=layer_slab,
        layers_contact1=layers_contact1,
        layers_contact2=layers_contact2,
        cladding_offsets_contact1=cladding_offsets_contact1,
        cladding_offsets_contact2=cladding_offsets_contact2,
        contact_width=contact_width,
        contact_gap=contact_gap,
        slab_gap=slab_gap,
        layer_via=layer_via,
        via_width=via_width,
        via_offsets=via_offsets,
    )
    x.info.update(**info)
    x.info.update(**kwargs)
    return x


@pydantic.validate_arguments
def pn(
    width: float = 0.5,
    layer: Tuple[int, int] = LAYER.WG,
    layer_slab: Tuple[int, int] = LAYER.SLAB90,
    gap_low_doping: float = 0.0,
    gap_medium_doping: Optional[float] = 0.5,
    gap_high_doping: Optional[float] = 1.0,
    width_doping: float = 8.0,
    width_slab: float = 7.0,
    layer_p: Tuple[int, int] = LAYER.P,
    layer_pp: Tuple[int, int] = LAYER.PP,
    layer_ppp: Tuple[int, int] = LAYER.PPP,
    layer_n: Tuple[int, int] = LAYER.N,
    layer_np: Tuple[int, int] = LAYER.NP,
    layer_npp: Tuple[int, int] = LAYER.NPP,
    cladding_offsets: Floats = (0,),
    layers_cladding: Optional[Tuple[Layer, ...]] = None,
    port_names: Tuple[str, str] = ("o1", "o2"),
) -> CrossSection:
    """rib PN doped cross_section.

    Args:
        width: width of the ridge
        layer: ridge llayer
        layer_slab: slab layer
        gap_low_doping: from waveguide center to low doping
        gap_medium_doping: from waveguide center to medium doping.
            None removes medium doping
        gap_high_doping: from waveguide center to high doping. None removes high doping
        width_doping:
        width_slab:
        layer_p:
        layer_pp:
        layer_ppp:
        layer_n:
        layer_np:
        layer_npp:
        cladding_offsets:
        layers_cladding: Iterable of layers
        port_names:


    .. code::

                                   layer
                           |<------width------>|
                            ____________________
                           |     |       |     |
        ___________________|     |       |     |__________________________|
                    P            |       |              N                 |
                 width_p         |       |           width_n              |
        <----------------------->|       |<------------------------------>|
                                     |<->|
                                     gap_low_doping
                                     |         |        N+                |
                                     |         |     width_np             |
                                     |         |<------------------------>|
                                     |<------->|
                                           gap_medium_doping

    """
    x = CrossSection()
    x.add(width=width, offset=0, layer=layer, ports=port_names)
    x.add(width=width_slab, offset=0, layer=layer_slab)

    offset_low_doping = width_doping / 2 + gap_low_doping
    width_low_doping = width_doping - gap_low_doping
    x.add(width=width_low_doping, offset=+offset_low_doping, layer=layer_n)
    x.add(width=width_low_doping, offset=-offset_low_doping, layer=layer_p)

    if gap_medium_doping is not None:
        width_medium_doping = width_doping - gap_medium_doping
        offset_medium_doping = width_medium_doping / 2 + gap_medium_doping

        x.add(
            width=width_medium_doping,
            offset=+offset_medium_doping,
            layer=layer_np,
        )
        x.add(
            width=width_medium_doping,
            offset=-offset_medium_doping,
            layer=layer_pp,
        )

    if gap_high_doping is not None:
        width_high_doping = width_doping - gap_high_doping
        offset_high_doping = width_high_doping / 2 + gap_high_doping
        x.add(width=width_high_doping, offset=+offset_high_doping, layer=layer_npp)
        x.add(width=width_high_doping, offset=-offset_high_doping, layer=layer_ppp)

    layers_cladding = layers_cladding or []

    for cladding_offset, layer_cladding in zip(cladding_offsets, layers_cladding):
        x.add(width=width_slab + 2 * cladding_offset, offset=0, layer=layer_cladding)

    s = dict(
        width=width,
        layer=layer,
        cladding_offsets=cladding_offsets,
        layers_cladding=layers_cladding,
        gap_low_doping=gap_low_doping,
        gap_medium_doping=gap_medium_doping,
        gap_high_doping=gap_high_doping,
        width_doping=width_doping,
        width_slab=width_slab,
    )
    x.info = s
    return x


@pydantic.validate_arguments
def strip_heater_metal_undercut(
    width: float = 0.5,
    layer: Layer = LAYER.WG,
    heater_width: float = 2.5,
    trench_width: float = 6.5,
    trench_gap: float = 2.0,
    layer_heater: Layer = LAYER.HEATER,
    layer_trench: Layer = LAYER.DEEPTRENCH,
    **kwargs,
) -> CrossSection:
    """Returns strip cross_section with top metal and undercut trenches on both sides.
    dimensions from https://doi.org/10.1364/OE.18.020298

    Args:
        width: of waveguide
        layer:
        heater_width: of metal heater
        trench_width:
        trench_gap: from waveguide edge to trench edge
        layer_heater:
        layer_trench:
        kwargs: cross_section settings


    .. code::

              |<-------heater_width--------->|
               ______________________________
              |                              |
              |         layer_heater         |
              |______________________________|

                   |<------width------>|
                    ____________________ trench_gap
                   |                   |<----------->|              |
                   |                   |             |   undercut   |
                   |       width       |             |              |
                   |                   |             |<------------>|
                   |___________________|             | trench_width |
                                                     |              |
                                                     |              |

    """
    trench_offset = trench_gap + trench_width / 2 + width / 2
    x = cross_section(
        width=width,
        layer=layer,
        sections=(
            Section(layer=layer_heater, width=heater_width),
            Section(layer=layer_trench, width=trench_width, offset=+trench_offset),
            Section(layer=layer_trench, width=trench_width, offset=-trench_offset),
        ),
        **kwargs,
    )
    x.info["heater_width"] = heater_width
    x.info["layer_heater"] = layer_heater
    x.info["trench_width"] = trench_width
    x.info["layer_heater"] = layer_heater
    x.info["trench_gap"] = trench_gap
    return x


@pydantic.validate_arguments
def strip_heater_metal(
    width: float = 0.5,
    layer: Layer = LAYER.WG,
    heater_width: float = 2.5,
    layer_heater: Layer = LAYER.HEATER,
    **kwargs,
) -> CrossSection:
    """Returns strip cross_section with top heater metal.
    dimensions from https://doi.org/10.1364/OE.18.020298

    Args:
        width: of waveguide
        layer:
        heater_width: of metal heater
        layer_heater: for the metal

    """
    x = cross_section(
        width=width,
        layer=layer,
        sections=(Section(layer=layer_heater, width=heater_width),),
        **kwargs,
    )
    x.info["heater_width"] = heater_width
    x.info["layer_heater"] = layer_heater
    return x


@pydantic.validate_arguments
def heater_metal(
    width: float = 2.5,
    layer: Layer = LAYER.HEATER,
    **kwargs,
) -> CrossSection:
    """Returns metal heater cross_section.
    dimensions from https://doi.org/10.1364/OE.18.020298

    Args:
        width: of the metal
        layer: of the heater

    """
    return cross_section(
        width=width,
        layer=layer,
        **kwargs,
    )


@pydantic.validate_arguments
def strip_heater_doped(
    width: float = 0.5,
    layer: Layer = LAYER.WG,
    heater_width: float = 2.0,
    heater_gap: float = 0.8,
    layers_heater: Layers = (LAYER.WG, LAYER.NPP),
    cladding_offsets_heater: Tuple[float, ...] = (0, 0.1),
    **kwargs,
) -> CrossSection:
    """Returns strip cross_section with N++ doped heaters on both sides.

    .. code::

                                  |<------width------>|
          ____________             ___________________               ______________
         |            |           |     undoped Si    |             |              |
         |layer_heater|           |  intrinsic region |<----------->| layer_heater |
         |____________|           |___________________|             |______________|
                                                                     <------------>
                                                        heater_gap     heater_width
    """
    heater_offset = width / 2 + heater_gap + heater_width / 2

    sections = [
        Section(
            layer=layer,
            width=heater_width + 2 * cladding_offset,
            offset=+heater_offset,
        )
        for layer, cladding_offset in zip(layers_heater, cladding_offsets_heater)
    ]

    sections += [
        Section(
            layer=layer,
            width=heater_width + 2 * cladding_offset,
            offset=-heater_offset,
        )
        for layer, cladding_offset in zip(layers_heater, cladding_offsets_heater)
    ]

    return cross_section(
        width=width,
        layer=layer,
        sections=sections,
        **kwargs,
    )


strip_heater_doped_contact = partial(
    strip_heater_doped,
    layers_heater=(LAYER.WG, LAYER.NPP, LAYER.VIAC),
    cladding_offsets_heater=(0, 0.1, -0.2),
)


@pydantic.validate_arguments
def rib_heater_doped(
    width: float = 0.5,
    layer: Layer = LAYER.WG,
    heater_width: float = 2.0,
    heater_gap: float = 0.8,
    layer_heater: Layer = LAYER.NPP,
    layer_slab: Layer = LAYER.SLAB90,
    slab_gap: float = 0.2,
    with_top_heater: bool = True,
    with_bot_heater: bool = True,
    **kwargs,
) -> CrossSection:
    """Returns rib cross_section with N++ doped heaters on both sides.
    dimensions from https://doi.org/10.1364/OE.27.010456

    .. code::


                                    |<------width------>|
                                     ____________________  heater_gap           slab_gap
                                    |                   |<----------->|             <-->
         ___ _______________________|                   |__________________________|___
        |   |            |                undoped Si                  |            |   |
        |   |layer_heater|                intrinsic region            |layer_heater|   |
        |___|____________|____________________________________________|____________|___|
                                                                       <---------->
                                                                        heater_width
        <------------------------------------------------------------------------------>
                                        slab_width
    """
    heater_offset = width / 2 + heater_gap + heater_width / 2

    if with_bot_heater and with_top_heater:
        slab_width = width + 2 * heater_gap + 2 * heater_width + 2 * slab_gap
        slab_offset = 0
    elif with_top_heater:
        slab_width = width + heater_gap + heater_width + slab_gap
        slab_offset = -slab_width / 2
    elif with_bot_heater:
        slab_width = width + heater_gap + heater_width + slab_gap
        slab_offset = +slab_width / 2

    sections = []

    if with_bot_heater:
        sections += [
            Section(layer=layer_heater, width=heater_width, offset=+heater_offset)
        ]
    if with_top_heater:
        sections += [
            Section(layer=layer_heater, width=heater_width, offset=-heater_offset)
        ]
    sections += [
        Section(width=slab_width, layer=layer_slab, offset=slab_offset, name="slab")
    ]
    return cross_section(
        width=width,
        layer=layer,
        sections=sections,
        **kwargs,
    )


@pydantic.validate_arguments
def rib_heater_doped_contact(
    width: float = 0.5,
    layer: Layer = LAYER.WG,
    heater_width: float = 1.0,
    heater_gap: float = 0.8,
    layer_slab: Layer = LAYER.SLAB90,
    layer_heater: Layer = LAYER.NPP,
    contact_width: float = 2.0,
    contact_gap: float = 0.8,
    layers_contact: Layers = (LAYER.NPP, LAYER.VIAC),
    cladding_offsets_contact: Tuple[float, ...] = (0, -0.2),
    slab_gap: float = 0.2,
    slab_offset: float = 0,
    with_top_heater: bool = True,
    with_bot_heater: bool = True,
    **kwargs,
) -> CrossSection:
    """Returns rib cross_section with N++ doped heaters on both sides.
    dimensions from https://doi.org/10.1364/OE.27.010456

    Args:
        width:
        layer:
        heater_width:
        heater_gap:
        layer_slab:
        layer_heater:
        contact_width:
        contact_gap:
        layers_contact:
        cladding_offsets_contact:
        slab_gap: from heater edge
        slab_offset: over the center of the slab
        with_top_heater:
        with_bot_heater:

    .. code::

                                   |<----width------>|
       slab_gap                     __________________ contact_gap     contact width
       <-->                        |                 |<------------>|<--------------->
                                   |                 | heater_gap |
                                   |                 |<---------->|
        ___ _______________________|                 |___________________________ ____
       |   |            |              undoped Si                 |              |    |
       |   |layer_heater|              intrinsic region           |layer_heater  |    |
       |___|____________|_________________________________________|______________|____|
                                                                   <------------>
                                                                    heater_width
       <------------------------------------------------------------------------------>
                                       slab_width

    """
    if with_bot_heater and with_top_heater:
        slab_width = width + 2 * heater_gap + 2 * heater_width + 2 * slab_gap
        slab_offset = slab_offset
    elif with_top_heater:
        slab_width = width + heater_gap + heater_width + slab_gap
        slab_offset = slab_offset - slab_width / 2
    elif with_bot_heater:
        slab_width = width + heater_gap + heater_width + slab_gap
        slab_offset = slab_offset + slab_width / 2

    heater_offset = width / 2 + heater_gap + heater_width / 2
    contact_offset = width / 2 + contact_gap + contact_width / 2
    sections = [
        Section(width=slab_width, layer=layer_slab, offset=slab_offset, name="slab"),
    ]
    if with_bot_heater:
        sections += [
            Section(
                layer=layer_heater,
                width=heater_width,
                offset=+heater_offset,
            )
        ]

    if with_top_heater:
        sections += [
            Section(
                layer=layer_heater,
                width=heater_width,
                offset=-heater_offset,
            )
        ]

    if with_bot_heater:
        sections += [
            Section(
                layer=layer,
                width=heater_width + 2 * cladding_offset,
                offset=+contact_offset,
            )
            for layer, cladding_offset in zip(layers_contact, cladding_offsets_contact)
        ]

    if with_top_heater:
        sections += [
            Section(
                layer=layer,
                width=heater_width + 2 * cladding_offset,
                offset=-contact_offset,
            )
            for layer, cladding_offset in zip(layers_contact, cladding_offsets_contact)
        ]

    return cross_section(
        sections=sections,
        width=width,
        layer=layer,
        **kwargs,
    )


strip = partial(cross_section)
strip_auto_widen = partial(cross_section, width_wide=0.9, auto_widen=True)
rib = partial(
    strip,
    sections=(Section(width=6, layer=LAYER.SLAB90, name="slab"),),
    layers_cladding=(LAYER.SLAB90,),
    cladding_offset=0,
)
nitride = partial(cross_section, layer=LAYER.WGN, width=1.0)
strip_rib_tip = partial(
    strip, sections=(Section(width=0.2, layer=LAYER.SLAB90, name="slab"),)
)

port_names_electrical = ("e1", "e2")
port_types_electrical = ("electrical", "electrical")
metal1 = partial(
    cross_section,
    layer=LAYER.M1,
    width=10.0,
    port_names=port_names_electrical,
    port_types=port_types_electrical,
)
metal2 = partial(
    metal1,
    layer=LAYER.M2,
    width=10.0,
    port_names=port_names_electrical,
    port_types=port_types_electrical,
)
metal3 = partial(
    metal1,
    layer=LAYER.M3,
    width=10.0,
    port_names=port_names_electrical,
    port_types=port_types_electrical,
)


# xs_strip = strip()
# xs_strip_auto_widen = strip_auto_widen()
# xs_rib = rib()
# xs_nitride = nitride()
# xs_metal1 = metal1()
# xs_metal2 = metal2()
# xs_metal3 = metal3()
# xs_pin = pin()
# xs_strip_heater_metal_undercut = strip_heater_metal_undercut()
# xs_strip_heater_metal = strip_heater_metal()
# xs_strip_heater_doped = strip_heater_doped()
# xs_rib_heater_doped = rib_heater_doped()


cross_section_factory = dict(
    cross_section=cross_section,
    strip=strip,
    strip_auto_widen=strip_auto_widen,
    rib=rib,
    nitride=nitride,
    metal1=metal1,
    metal2=metal2,
    metal3=metal3,
    pin=pin,
    strip_heater_metal_undercut=strip_heater_metal_undercut,
    strip_heater_metal=strip_heater_metal,
    strip_heater_doped=strip_heater_doped,
    rib_heater_doped=rib_heater_doped,
)

if __name__ == "__main__":
    import gdsfactory as gf

    P = gf.path.straight()
    # P = gf.path.euler(radius=10, use_eff=True)
    # P = euler()
    # P = gf.Path()
    # P.append(gf.path.straight(length=5))
    # P.append(gf.path.arc(radius=10, angle=90))
    # P.append(gf.path.spiral())

    # Create a blank CrossSection
    # X.add(width=2.0, offset=4, layer=LAYER.HEATER, ports=["HW0", "HE0"])

    # X = pin(width=0.5, width_i=0.5)
    # x = strip(width=0.5)

    # X = strip_heater_metal_undercut()
    # X = metal1()
    # X = pin(layer_via=LAYER.VIAC, via_offsets=(-2, 2))
    # X = pin()
    # X = strip_heater_doped()

    # x1 = strip_rib_tip()
    # x2 = rib_heater_doped_contact()
    # X = gf.path.transition(x1, x2)
    # P = gf.path.straight(npoints=100, length=10)

    # X = CrossSection()
    # X.add(width=2.0, offset=-4, layer=LAYER.HEATER, ports=["e1", "e2"])
    # X.add(width=0.5, offset=0, layer=LAYER.SLAB90, ports=["o1", "o2"])

    X = rib_heater_doped(with_bot_heater=False)
    P = gf.path.straight(npoints=100, length=10)

    c = gf.path.extrude(P, X)

    # print(x1.to_dict())
    # print(x1.name)

    # c = gf.path.component(P, strip(width=2, layer=LAYER.WG, cladding_offset=3))
    # c = gf.add_pins(c)
    # c << gf.components.bend_euler(radius=10)
    # c << gf.components.bend_circular(radius=10)
    # c.pprint_ports()
    c.show()



import pytest

import gdsfactory as gf
from gdsfactory.routing.manhattan import RouteWarning


def test_route_error_bundle():
    """Ensures that an impossible route raises value Error"""
    c = gf.Component("get_route_from_steps_sample")

    w = gf.components.array(
        gf.partial(gf.components.straight, layer=(2, 0)),
        rows=3,
        columns=1,
        spacing=(0, 50),
    )

    left = c << w
    right = c << w
    right.move((200, 100))
    p1 = left.get_ports_list(orientation=0)
    p2 = right.get_ports_list(orientation=180)

    with pytest.warns(RouteWarning):
        routes = gf.routing.get_bundle_from_steps(
            p1,
            p2,
            steps=[{"x": 300}, {"x": 301}],
        )

    for route in routes:
        c.add(route.references)
    return c


if __name__ == "__main__":
    c = test_route_error_bundle()
    c.show()



import itertools as it

import jsondiff
import numpy as np
import pytest
from omegaconf import OmegaConf
from pytest_regressions.data_regression import DataRegressionFixture

from gdsfactory.component import Component
from gdsfactory.difftest import difftest
from gdsfactory.read.from_yaml import from_yaml, sample_mmis

sample_connections = """
name: sample_connections

instances:
    wgw:
      component: straight
      settings:
        length: 1
    wgn:
      component: straight
      settings:
        length: 0.5

connections:
    wgw,o1: wgn,o2

"""

#
#        __Lx__
#       |      |
#       Ly     Lyr
#       |      |
#  CP1==|      |==CP2
#       |      |
#       Ly     Lyr
#       |      |
#      DL/2   DL/2
#       |      |
#       |__Lx__|
#


sample_mirror_simple = """
name: sample_mirror_simple

instances:
    s:
        component: straight

    b:
        component: bend_circular

placements:
    b:
        mirror: True
        port: o1

connections:
    b,o1: s,o2

"""


def test_sample() -> Component:
    c = from_yaml(sample_mmis)
    assert len(c.get_dependencies()) == 6, len(c.get_dependencies())
    assert len(c.ports) == 3, len(c.ports)
    return c


def test_connections() -> Component:
    c = from_yaml(sample_connections)
    assert len(c.get_dependencies()) == 2
    assert len(c.ports) == 0
    return c


sample_2x2_connections = """
name: connections_2x2_solution

instances:
    mmi_bottom:
      component: mmi2x2
      settings:
            length_mmi: 5
    mmi_top:
      component: mmi2x2
      settings:
            length_mmi: 10

placements:
    mmi_top:
        x: 100
        y: 100

routes:
    optical:
        links:
            mmi_bottom,o4: mmi_top,o1
            mmi_bottom,o3: mmi_top,o2

        settings:
            layer: [2, 0]

"""


def test_connections_2x2() -> Component:
    c = from_yaml(sample_2x2_connections)
    assert len(c.get_dependencies()) == 8, len(c.get_dependencies())
    assert len(c.ports) == 0, len(c.ports)

    length = c.routes["mmi_bottom,o3:mmi_top,o2"]
    assert np.isclose(length, 165.774), length
    return c


sample_different_factory = """
name: sample_different_factory

instances:
    bl:
      component: pad
    tl:
      component: pad
    br:
      component: pad
    tr:
      component: pad

placements:
    tl:
        x: 0
        y: 200

    br:
        x: 400
        y: 400

    tr:
        x: 400
        y: 600

routes:
    electrical:
        settings:
            separation: 20
            layer: [31, 0]
            width: 10
        links:
            tl,e3: tr,e1
            bl,e3: br,e1
    optical:
        settings:
            radius: 100
        links:
            bl,e4: br,e3

"""


def test_connections_different_factory() -> Component:
    c = from_yaml(sample_different_factory)
    lengths = [693.274, 693.274, 1199.144]

    assert np.isclose(c.routes["tl,e3:tr,e1"], lengths[0]), c.routes["tl,e3:tr,e1"]
    assert np.isclose(c.routes["bl,e3:br,e1"], lengths[1]), c.routes["bl,e3:br,e1"]
    assert np.isclose(c.routes["bl,e4:br,e3"], lengths[2]), c.routes["bl,e4:br,e3"]

    return c


sample_different_link_factory = """
name: sample_path_length_matching

instances:
    bl:
      component: pad
    tl:
      component: pad
    br:
      component: pad
    tr:
      component: pad

placements:
    tl:
        x: 0
        y: 200

    br:
        x: 900
        y: 400

    tr:
        x: 900
        y: 600

routes:
    route1:
        routing_strategy: get_bundle_path_length_match
        settings:
            radius: 10
            extra_length: 500
        links:
            tl,e3: tr,e1
            bl,e3: br,e1

"""


def test_connections_different_link_factory() -> Component:
    c = from_yaml(sample_different_link_factory)

    length = 1719.822
    assert np.isclose(c.routes["tl,e3:tr,e1"], length), c.routes["tl,e3:tr,e1"]
    assert np.isclose(c.routes["bl,e3:br,e1"], length), c.routes["bl,e3:br,e1"]
    return c


sample_waypoints = """
name: sample_waypoints

instances:
    t:
      component: pad_array
      settings:
          orientation: 270
    b:
      component: pad_array
      settings:
          orientation: 90

placements:
    t:
        x: -250
        y: 1000
routes:
    route1:
        routing_strategy: get_bundle_from_waypoints
        settings:
            waypoints:
                - [0, 300]
                - [400, 300]
                - [400, 400]
                - [-250, 400]
            auto_widen: False
        links:
            b,e11: t,e11
            b,e12: t,e12
"""


sample_docstring = """
name: sample_docstring

instances:
    mmi_bot:
      component: mmi1x2
      settings:
        width_mmi: 5
        length_mmi: 11
    mmi_top:
      component: mmi1x2
      settings:
        width_mmi: 6
        length_mmi: 22

placements:
    mmi_top:
        port: o1
        x: 0
        y: 0
    mmi_bot:
        port: o1
        x: mmi_top,o2
        y: mmi_top,o2
        dx: 40
        dy: -40
routes:
    optical:
        links:
            mmi_top,o3: mmi_bot,o1
"""


sample_regex_connections = """
name: sample_regex_connections

instances:
    left:
      component: nxn
      settings:
        west: 0
        east: 3
        ysize: 20
    right:
      component: nxn
      settings:
        west: 3
        east: 0
        ysize: 20

placements:
    right:
        x: 20
routes:
    optical:
        links:
            left,o:1:3: right,o:1:3
"""

sample_regex_connections_backwards = """
name: sample_regex_connections_backwards

instances:
    left:
      component: nxn
      settings:
        west: 0
        east: 3
        ysize: 20
    right:
      component: nxn
      settings:
        west: 3
        east: 0
        ysize: 20

placements:
    right:
        x: 20
routes:
    optical:
        links:
            left,o:3:1: right,o:3:1
"""


def test_connections_regex() -> Component:
    c = from_yaml(sample_regex_connections)
    route_names = ["left,o1:right,o1", "left,o2:right,o2", "left,o3:right,o3"]

    length = 12.0
    for route_name in route_names:
        assert np.isclose(c.routes[route_name], length)
    return c


def test_connections_regex_backwargs() -> Component:
    c = from_yaml(sample_regex_connections_backwards)
    route_names = ["left,o1:right,o1", "left,o2:right,o2", "left,o3:right,o3"]

    length = 12.0
    for route_name in route_names:
        assert np.isclose(c.routes[route_name], length), c.routes[route_name]
    return c


def test_connections_waypoints() -> Component:
    c = from_yaml(sample_waypoints)

    length = 2036.548
    route_name = "b,e11:t,e11"
    assert np.isclose(c.routes[route_name], length), c.routes[route_name]
    return c


def test_docstring_sample() -> Component:
    c = from_yaml(sample_docstring)
    route_name = "mmi_top,o3:mmi_bot,o1"
    length = 72.024
    assert np.isclose(c.routes[route_name], length), c.routes[route_name]
    return c


yaml_fail = """
name: yaml_fail
instances:
    mmi_long:
      component: mmi1x2
      settings:
        width_mmi: 4.5
        length_mmi: 10
    mmi_short:
      component: mmi1x2
      settings:
        width_mmi: 4.5
        length_mmi: 5

placements:
    mmi_short:
        port: o1
        x: mmi_long,o2
        y: mmi_long,o2
    mmi_long:
        port: o1
        x: mmi_short,o2
        y: mmi_short,o2
        dx : 10
        dy: 20
"""
#                      ______
#                     |      |
#           dx   1----| short|
#                |    |______|
#                | dy
#   ______ north |
#   |     |
#   |long |
#   |_____|
#       east

yaml_anchor = """
name: yaml_anchor
instances:
    mmi_long:
      component: mmi1x2
      settings:
        width_mmi: 4.5
        length_mmi: 10
    mmi_short:
      component: mmi1x2
      settings:
        width_mmi: 4.5
        length_mmi: 5

placements:
    mmi_short:
        port: o3
        x: 0
        y: 0
    mmi_long:
        port: o1
        x: mmi_short,east
        y: mmi_short,north
        dx : 10
        dy: 10
"""

# FIXME: Fix both unconmmented cases
# yaml_fail should actually fail
# sample_different_factory: returns a zero length straight that gives an error
# when extracting the netlist

yaml_strings = dict(
    yaml_anchor=yaml_anchor,
    # yaml_fail=yaml_fail,
    # sample_regex_connections_backwards=sample_regex_connections_backwards,
    # sample_regex_connections=sample_regex_connections,
    sample_docstring=sample_docstring,
    sample_waypoints=sample_waypoints,
    sample_different_link_factory=sample_different_link_factory,
    # sample_different_factory=sample_different_factory,
    sample_mirror_simple=sample_mirror_simple,
    sample_connections=sample_connections,
    sample_mmis=sample_mmis,
)


@pytest.mark.parametrize("yaml_key", yaml_strings.keys())
def test_gds(yaml_key: str, data_regression: DataRegressionFixture) -> None:
    """Avoid regressions in GDS geometry shapes and layers."""
    yaml_string = yaml_strings[yaml_key]
    c = from_yaml(yaml_string)
    difftest(c)


@pytest.mark.parametrize("yaml_key", yaml_strings.keys())
def test_settings(
    yaml_key: str, data_regression: DataRegressionFixture, check: bool = True
) -> Component:
    """Avoid regressions when exporting settings."""
    yaml_string = yaml_strings[yaml_key]
    c = from_yaml(yaml_string)

    if check:
        data_regression.check(c.to_dict())
    return c


@pytest.mark.parametrize(
    "yaml_key,full_settings", it.product(yaml_strings.keys(), [True, False])
)
def test_netlists(
    yaml_key: str,
    full_settings: bool,
    data_regression: DataRegressionFixture,
    check: bool = True,
) -> None:
    """Write netlists for hierarchical circuits.
    Checks that both netlists are the same
    jsondiff does a hierarchical diff

    Component -> netlist -> Component -> netlist
    """
    yaml_string = yaml_strings[yaml_key]
    c = from_yaml(yaml_string)
    n = c.get_netlist(full_settings=full_settings)
    if check:
        data_regression.check(OmegaConf.to_container(n))

    yaml_str = OmegaConf.to_yaml(n, sort_keys=True)
    # print(yaml_str)
    c2 = from_yaml(yaml_str)
    n2 = c2.get_netlist(full_settings=full_settings)
    d = jsondiff.diff(n, n2)
    assert len(d) == 0, print(d)
    return c2


def _demo_netlist():
    """path on the route"""
    import gdsfactory as gf

    # c = from_yaml(sample_2x2_connections)
    c = from_yaml(sample_waypoints)
    c = from_yaml(sample_different_factory)
    c.show()
    full_settings = True
    n = c.get_netlist(full_settings=full_settings)
    yaml_str = OmegaConf.to_yaml(n, sort_keys=True)
    c2 = from_yaml(yaml_str)
    n2 = c2.get_netlist(full_settings=full_settings)
    d = jsondiff.diff(n, n2)
    assert len(d) == 0
    gf.show(c2)


if __name__ == "__main__":
    # c = from_yaml(sample_2x2_connections)
    # c = from_yaml(sample_different_factory)
    # c = test_sample()
    # c = test_netlists("sample_mmis", True, None, check=False)
    # c = test_connections_regex()
    # c = test_connections_regex_backwargs()
    # c = test_mirror()
    # c = test_connections()
    # c = test_connections_different_factory()

    # c = test_sample()
    # c = test_connections_2x2()
    # c = test_connections_different_factory()
    # c = test_connections_different_link_factory()
    # c = test_connections_waypoints()
    # c = test_docstring_sample()
    # c = test_settings("yaml_anchor", None, False)
    # c = test_netlists("yaml_anchor", True, None, False)
    # c = test_netlists("sample_waypoints", True, None, False)
    # c = from_yaml(sample_docstring)
    # c = from_yaml(sample_different_link_factory)
    # c = from_yaml(sample_mirror_simple)
    # c = from_yaml(sample_waypoints)
    c = test_netlists("sample_different_link_factory", True, None, check=False)

    # c = from_yaml(sample_different_factory)
    # c = from_yaml(sample_different_link_factory)
    # c = from_yaml(sample_waypoints)
    # c = from_yaml(sample_docstring)
    # c = from_yaml(sample_regex_connections)
    # c = from_yaml(sample_regex_connections_backwards)
    c.show()



import warnings
from typing import Optional, Tuple, Union

import numpy as np
from numpy import ndarray

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.mmi1x2 import mmi1x2
from gdsfactory.cross_section import cross_section as cross_section_function
from gdsfactory.port import Port
from gdsfactory.types import ComponentOrFactory, Coordinate, CrossSectionFactory, Layer

DEG2RAD = np.pi / 180


def line(
    p_start: Union[Port, Coordinate],
    p_end: Union[Port, Coordinate],
    width: Optional[float] = None,
) -> Tuple[float, float, float, float]:
    if isinstance(p_start, gf.Port):
        width = p_start.width
        p_start = p_start.midpoint

    if isinstance(p_end, gf.Port):
        p_end = p_end.midpoint

    w = width
    angle = np.arctan2(p_end[1] - p_start[1], p_end[0] - p_start[0])
    a = np.pi / 2
    p0 = move_polar_rad_copy(p_start, angle + a, w / 2)
    p1 = move_polar_rad_copy(p_start, angle - a, w / 2)
    p2 = move_polar_rad_copy(p_end, angle - a, w / 2)
    p3 = move_polar_rad_copy(p_end, angle + a, w / 2)
    return p0, p1, p2, p3


def move_polar_rad_copy(pos: Coordinate, angle: float, length: float) -> ndarray:
    """Returns the points of a position (pos) with angle, by shifted by certain length

    Args:
        pos: position
        angle: in radians
        length: extension length

    """
    c = np.cos(angle)
    s = np.sin(angle)
    return pos + length * np.array([c, s])


@cell
def extend_port(port: Port, length: float, layer: Optional[Layer] = None) -> Component:
    """Returns a straight extension component out of a port.

    Args:
        port: port to extend
        length: extension length
        layer: for the straight section
    """
    c = Component()
    layer = layer or port.layer

    # Generate a port extension
    p_start = port.midpoint
    angle = port.angle
    p_end = move_polar_rad_copy(p_start, angle * DEG2RAD, length)
    w = port.width

    _line = line(p_start, p_end, w)

    c.add_polygon(_line, layer=layer)
    c.add_port(name="original", port=port)

    port_settings = port.settings.copy()
    port_settings.update(midpoint=p_end)
    c.add_port(**port_settings)

    return c


@gf.cell
def extend_ports(
    component: ComponentOrFactory = mmi1x2,
    port_names: Optional[Tuple[str, ...]] = None,
    length: float = 5.0,
    extension_factory: Optional[ComponentOrFactory] = None,
    port1: Optional[str] = None,
    port2: Optional[str] = None,
    port_type: str = "optical",
    centered: bool = False,
    cross_section: Optional[CrossSectionFactory] = None,
    **kwargs,
) -> Component:
    """Returns a new component with some ports extended
    you can define an extension_factory
    otherwise, defaults to the port cross_section of each port to extend

    Args:
        component: component to extend ports
        port_names: list of ports names to extend, if None it extends all ports
        length: extension length
        extension_factory: function to extend ports (defaults to a straight)
        port1: input port name
        port2: output port name
        port_type: type of the ports to extend
        centered: if True centers rectangle at (0, 0)
        cross_section: extension cross_section, defaults to port cross_section
            if port has no cross_section it creates one using width and layer

    Keyword Args:
        layer: port GDS layer
        prefix: port name prefix
        orientation: in degrees
        width: port width
        layers_excluded: List of layers to exclude
        port_type: optical, electrical, ...
        clockwise: if True, sort ports clockwise, False: counter-clockwise
    """
    c = gf.Component()
    component = component() if callable(component) else component
    cref = c << component
    c.component = component

    if centered:
        cref.x = 0
        cref.y = 0

    ports_all = cref.get_ports_list()
    port_names_all = [p.name for p in ports_all]

    ports_to_extend = cref.get_ports_list(port_type=port_type, **kwargs)
    ports_to_extend_names = [p.name for p in ports_to_extend]
    ports_to_extend_names = port_names or ports_to_extend_names or port_names_all

    for port_name in ports_to_extend_names:
        if port_name not in port_names_all:
            warnings.warn(f"Port Name {port_name} not in {port_names_all}")

    for port in ports_all:
        port_name = port.name
        port = cref.ports[port_name]
        cross_section_extension = (
            cross_section
            or port.cross_section
            or cross_section_function(layer=port.layer, width=port.width)
        )

        if port_name in ports_to_extend_names:
            if extension_factory:
                extension_component = extension_factory()
            else:
                extension_component = gf.partial(
                    gf.components.straight,
                    length=length,
                    width=port.width,
                    cross_section=cross_section_extension,
                    layer=port.layer,
                )()
            port_labels = list(extension_component.ports.keys())
            port1 = port1 or port_labels[0]
            port2 = port2 or port_labels[-1]

            extension = c << extension_component
            extension.connect(port1, port)
            c.add_port(port_name, port=extension.ports[port2])
            c.absorb(extension)
        else:
            c.add_port(port_name, port=component.ports[port_name])

    c.copy_child_info(component)
    return c


def test_extend_ports() -> Component:
    import gdsfactory as gf
    import gdsfactory.components as pc

    width = 0.5
    xs_strip = gf.partial(gf.cross_section.strip, width=width)

    c = pc.cross(width=width, port_type="optical")

    c1 = extend_ports(component=c, cross_section=xs_strip)
    assert len(c.ports) == len(c1.ports)
    p = len(c1.polygons)
    assert p == 4, p

    c2 = extend_ports(component=c, cross_section=xs_strip, port_names=("o1", "o2"))
    p = len(c2.polygons)
    assert p == 2, p

    c3 = extend_ports(component=c, cross_section=xs_strip)
    p = len(c3.polygons)
    assert p == 4, p

    c4 = extend_ports(gf.components.cross(port_type="optical"))
    p = len(c4.polygons)
    assert p == 4, p

    return c4


__all__ = ["extend_ports", "extend_port"]


if __name__ == "__main__":
    # c = extend_ports()
    # c = extend_ports(gf.components.mzi_phase_shifter_top_heater_metal)
    c = test_extend_ports()

    # c = extend_ports(gf.components.cross(port_type="optical"))
    c.show()

    # c = gf.components.bend_circular()
    # ce = extend_ports(component=c, port_names=list(c.ports.keys()) + ["hi"])
    # ce.show()

    # c = gf.components.straight_pin(length=40)
    # ce = extend_ports(c, port_names=("o1", "o2"))
    # ce = extend_ports(c, port_names=("top_e1", "bot_e3"))
    # ce = extend_ports(c, port_names=("bad", "worse"))
    # ce.show()
    # wg_pin.show()

    # c = pc.straight(layer=(3, 0))
    # print(ce)
    # print(len(ce.ports))
    # c = pc.straight()
    # ce = extend_ports(component=c)
    # ce.show()



from gdsfactory.samples.pdk.fab_e.schema import (
    CrossSectionModel,
    SectionModel,
    TechModel,
)

layers = dict(wg=(1, 0), wg_clad=(1, 2), pin=(100, 0), heater=(11, 0))
xs_strip = CrossSectionModel()
xs_strip_heater_metal = CrossSectionModel(
    width=0.5, sections=[SectionModel(layer=layers["heater"], width=2.5)]
)
cross_sections = dict(xs_strip=xs_strip, xs_strip_heater_metal=xs_strip_heater_metal)
TECH = TechModel(cross_sections=cross_sections, layers=layers)


if __name__ == "__main__":
    print(xs_strip_heater_metal)



"""Picwriter is a photonics library written by Derek Kita.

https://picwriter.readthedocs.io/en/latest/component-documentation.html
As it is based on gdspy it's easier to wrap picwriter components
"""
import gdspy
import numpy as np
import picwriter.components as pc
import picwriter.toolkit as pt
from picwriter.components.waveguide import WaveguideTemplate
from picwriter.toolkit import Component

import gdsfactory as gf
from gdsfactory.types import CrossSectionFactory, Layer

gdspy.current_library = gdspy.GdsLibrary()

direction_to_orientation = dict(EAST=0.0, NORTH=90.0, WEST=180.0, SOUTH=270.0)


def cross_section_to_waveguide_template(
    cross_section: CrossSectionFactory,
    euler_bend: bool = True,
    wg_type: str = "strip",
    **kwargs
) -> WaveguideTemplate:
    x = cross_section(**kwargs)

    layer = x.info["layer"]
    layer_cladding = x.info["layers_cladding"][0]

    return pc.WaveguideTemplate(
        bend_radius=x.info["radius"],
        wg_width=x.info["width"],
        wg_layer=layer[0],
        wg_datatype=layer[1],
        clad_layer=layer_cladding[0],
        clad_datatype=layer_cladding[1],
        clad_width=x.info["cladding_offset"],
        wg_type=wg_type,
        euler_bend=euler_bend,
    )


def direction_to_degree(direction: str) -> float:
    """Convert a 'direction' (as used in picwriter) to an angle in degrees.
    picwriter 'direction's can be either a float (corresponding to an angle in radians)
    or a string, corresponding to a cardinal direction
    """
    if isinstance(direction, float):
        # direction is a float in radians, but rotation should be a float in degrees
        return direction * 180.0 / np.pi
    return direction_to_orientation[direction]


def from_picwriter(
    picwriter_object: pt.Component, port_layer: Layer = (1, 0)
) -> Component:
    """Convert a Picwriter into a Gdsfactory Component."""
    po = picwriter_object
    c = gf.Component(name=po.name_prefix)

    # Extract the relevant cells from the picwriter global cell list
    po_cell = pt.CURRENT_CELLS[po.cell_hash]

    polygons = po_cell.get_polygonsets()
    for poly in polygons:
        polygons = poly.polygons
        layers = poly.layers
        datatypes = poly.datatypes

        for polygon, layer, datatype in zip(polygons, layers, datatypes):
            c.add_polygon(polygon, layer=(layer, datatype))

    c2 = gf.Component()
    ref = c2.add_ref(c)

    translate_by = po.port
    rotate_by = direction_to_degree(po.direction)
    ref.rotate(rotate_by)  # First rotate about (0,0)
    ref.move(translate_by)  # Next translate

    for port in po.portlist.keys():
        port_loc = po.portlist[port]["port"]
        direction = direction_to_degree(po.portlist[port]["direction"])

        c2.add_port(
            name=port,
            midpoint=[port_loc[0], port_loc[1]],
            width=po.wgt.wg_width,
            orientation=direction,
            layer=port_layer,
        )

    c2.absorb(ref)
    c2.auto_rename_ports()
    return c2


if __name__ == "__main__":

    wgt = pc.WaveguideTemplate(
        bend_radius=50.0,
        wg_width=1.0,
        wg_layer=1,
        wg_datatype=0,
        clad_layer=2,
        clad_datatype=0,
    )

    # gc = pc.GratingCoupler(wgt, port=(10, 20), direction=np.pi * 7 / 8)
    gc = pc.GratingCoupler(wgt, port=(10, 20), direction=0.0)
    gcc = from_picwriter(gc)

    gf.show(gcc)



import pytest

import gdsfactory as gf
from gdsfactory.samples.pdk.fab_c import (
    LAYER,
    WIDTH_NITRIDE_CBAND,
    bend_euler_c,
    straight_c,
    xs_nitridec,
)


@pytest.mark.parametrize("optical_routing_type", [0, 1])
def test_add_pins_with_routes(optical_routing_type) -> gf.Component:
    """
    Add grating couplers to a straight
    ensure that all the routes have pins

    """
    c = straight_c(length=11.0)
    gc = gf.components.grating_coupler_elliptical_te(
        wg_width=WIDTH_NITRIDE_CBAND, layer=LAYER.WGN
    )
    cc = gf.routing.add_fiber_single(
        component=c,
        grating_coupler=[gc, gf.components.grating_coupler_tm],
        cross_section=xs_nitridec,
        straight=straight_c,
        bend=bend_euler_c,
        optical_routing_type=optical_routing_type,
    )
    pins_component = cc.extract(layers=(LAYER.PIN,))
    pins_component.name = "test_add_pins_with_routes_component"
    assert len(pins_component.polygons) == 8, len(pins_component.polygons)
    return cc


def test_add_pins() -> None:
    """ensure that all the waveguide has 2 pins"""
    c = straight_c(length=11.0)
    pins_component = c.extract(layers=(LAYER.PIN,))
    pins_component.name = "test_add_pins_component"
    assert len(pins_component.polygons) == 2, len(pins_component.polygons)


if __name__ == "__main__":
    # test_add_pins()
    c = test_add_pins_with_routes(0)
    c.show()

    # test_add_pins_with_routes(1)

    # c = straight_c()
    # gc = gf.components.grating_coupler_elliptical_te(wg_width=WIDTH_NITRIDE_CBAND)
    # cc = gf.routing.add_fiber_single(
    #     component=c,
    #     grating_coupler=gc,
    #     straight=straight_c,
    #     optical_routing_type=1,
    # )
    # cc.show()
    # pins_component = cc.extract(layers=(LAYER.PIN,))
    # print(len(pins_component.polygons))



from gdsfactory.add_loopback import straight_with_loopback
from gdsfactory.component import Component
from gdsfactory.difftest import difftest


def test_add_loopback() -> Component:
    c = straight_with_loopback()
    difftest(c)
    return c


if __name__ == "__main__":
    c = test_add_loopback()
    c.show()



import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.config import TECH
from gdsfactory.cross_section import strip
from gdsfactory.path import arc, extrude
from gdsfactory.snap import snap_to_grid
from gdsfactory.types import CrossSectionFactory


@gf.cell
def bend_circular_heater(
    radius: float = 10,
    angle: int = 90,
    npoints: int = 720,
    heater_to_wg_distance: float = 1.2,
    heater_width: float = 0.5,
    layer_heater=TECH.layer.HEATER,
    cross_section: CrossSectionFactory = strip,
) -> Component:
    """Creates an arc of arclength ``theta`` starting at angle ``start_angle``

    Args:
        radius
        angle: angle of arc (degrees)
        npoints: Number of points used per 360 degrees
        heater_to_wg_distance:
        heater_width
        layer_heater
        cross_section:
    """
    x = cross_section()
    width = x.info["width"]
    cladding_offset = x.info["cladding_offset"]
    layers_cladding = x.info["layers_cladding"] or []
    layer = x.info["layer"]

    x = gf.CrossSection()
    x.add(width=width, offset=0, layer=layer, ports=["in", "out"])

    for layer_cladding in layers_cladding:
        x.add(width=width + 2 * cladding_offset, offset=0, layer=layer_cladding)

    offset = heater_to_wg_distance + width / 2
    x.add(
        width=heater_width,
        offset=+offset,
        layer=layer_heater,
    )
    x.add(
        width=heater_width,
        offset=-offset,
        layer=layer_heater,
    )
    p = arc(radius=radius, angle=angle, npoints=npoints)
    c = extrude(p, x)
    c.length = snap_to_grid(p.length())
    c.dx = abs(p.points[0][0] - p.points[-1][0])
    c.dy = abs(p.points[0][0] - p.points[-1][0])
    return c


if __name__ == "__main__":
    c = bend_circular_heater(heater_width=1)
    print(c.ports)
    c.show()



"""In programming, a factory is a function that returns an Object.

Functions are easy to understand because they have clear inputs and outputs.
Most gdsfactory functions take some inputs and return a Component object.
Some of these inputs parameters are also functions.

- Component: Object with.
    - name
    - references to other components (x, y, rotation)
    - polygons in different layers
    - ports dictionary
- ComponentFactory: function that returns a Component.
- Route: dataclass with 3 attributes.
    - references: list of references (straights, bends and tapers)
    - ports: dict(input=PortIn, output=PortOut)
    - length: float (how long is this route)
- RouteFactory: function that returns a Route.

"""
import json
import pathlib
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from omegaconf import OmegaConf
from phidl.device_layout import Label as LabelPhidl
from phidl.device_layout import Path
from pydantic import BaseModel
from typing_extensions import Literal

from gdsfactory.component import Component, ComponentReference
from gdsfactory.cross_section import CrossSection
from gdsfactory.port import Port

Anchor = Literal[
    "ce",
    "cw",
    "nc",
    "ne",
    "nw",
    "sc",
    "se",
    "sw",
    "center",
    "cc",
]
Axis = Literal["x", "y"]

NSEW = Literal["N", "S", "E", "W"]


class Label(LabelPhidl):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        """check with pydantic Label valid type"""
        assert isinstance(v, LabelPhidl), f"TypeError, Got {type(v)}, expecting Label"
        return v


class Route(BaseModel):
    references: List[ComponentReference]
    labels: Optional[List[Label]] = None
    ports: Tuple[Port, Port]
    length: float


class Routes(BaseModel):
    references: List[ComponentReference]
    lengths: List[float]
    ports: Optional[List[Port]] = None
    bend_radius: Optional[List[float]] = None


class ComponentModel(BaseModel):
    component: str
    settings: Optional[Dict[str, Any]]


class PlacementModel(BaseModel):
    x: Union[str, float] = 0
    y: Union[str, float] = 0
    dx: float = 0
    dy: float = 0
    port: Optional[Union[str, Anchor]] = None
    rotation: int = 0
    mirror: bool = False


class RouteModel(BaseModel):
    links: Dict[str, str]
    settings: Optional[Dict[str, Any]] = None
    routing_strategy: Optional[str] = None


class CircuitModel(BaseModel):
    instances: Dict[str, ComponentModel]
    name: Optional[str] = None
    placements: Optional[Dict[str, PlacementModel]] = None
    connections: Optional[List[Dict[str, str]]] = None
    routes: Optional[Dict[str, RouteModel]] = None
    info: Optional[Dict[str, Any]] = None


Float2 = Tuple[float, float]
Float3 = Tuple[float, float, float]
Floats = Tuple[float, ...]
Strs = Tuple[str, ...]
Int2 = Tuple[int, int]
Int3 = Tuple[int, int, int]
Ints = Tuple[int, ...]

Layer = Tuple[int, int]
Layers = Tuple[Layer, ...]
RouteFactory = Callable[..., Route]
ComponentFactory = Callable[..., Component]
ComponentFactoryDict = Dict[str, ComponentFactory]
PathFactory = Callable[..., Path]
PathType = Union[str, pathlib.Path]
PathTypes = Tuple[PathType, ...]

ComponentOrFactory = Union[ComponentFactory, Component]
ComponentOrFactoryOrList = Union[ComponentOrFactory, List[ComponentOrFactory]]
ComponentOrPath = Union[PathType, Component]
ComponentOrReference = Union[Component, ComponentReference]
NameToFunctionDict = Dict[str, ComponentFactory]
Number = Union[float, int]
Coordinate = Tuple[float, float]
Coordinates = Tuple[Coordinate, ...]
ComponentOrPath = Union[Component, PathType]
CrossSectionFactory = Callable[..., CrossSection]
CrossSectionOrFactory = Union[CrossSection, Callable[..., CrossSection]]
PortSymmetries = Dict[str, Dict[str, List[str]]]


__all__ = (
    "ComponentFactory",
    "ComponentFactoryDict",
    "ComponentOrFactory",
    "ComponentOrPath",
    "ComponentOrReference",
    "Coordinate",
    "Coordinates",
    "CrossSectionFactory",
    "CrossSectionOrFactory",
    "Float2",
    "Float3",
    "Floats",
    "Int2",
    "Int3",
    "Ints",
    "Layer",
    "Layers",
    "NameToFunctionDict",
    "Number",
    "PathType",
    "PathTypes",
    "Route",
    "RouteFactory",
    "Routes",
    "Strs",
)


def write_schema(model: BaseModel = CircuitModel):
    s = model.schema_json()
    d = OmegaConf.create(s)

    f1 = pathlib.Path(__file__).parent / "schema.yaml"
    f1.write_text(OmegaConf.to_yaml(d))

    f2 = pathlib.Path(__file__).parent / "schema.json"
    f2.write_text(json.dumps(OmegaConf.to_container(d)))


if __name__ == "__main__":
    write_schema()



"""Straight waveguide."""
import gdsfactory as gf
from gdsfactory.add_padding import get_padding_points
from gdsfactory.component import Component
from gdsfactory.cross_section import strip
from gdsfactory.types import CrossSectionOrFactory


@gf.cell
def straight(
    length: float = 10.0,
    npoints: int = 2,
    with_cladding_box: bool = True,
    cross_section: CrossSectionOrFactory = strip,
    **kwargs
) -> Component:
    """Returns a Straight waveguide.

    Args:
        length: straight length
        npoints: number of points
        with_cladding_box: box in layers_cladding to avoid DRC sharp edges
        cross_section:
        **kwargs: cross_section settings
    """
    p = gf.path.straight(length=length, npoints=npoints)
    x = cross_section(**kwargs) if callable(cross_section) else cross_section

    c = Component()
    path = gf.path.extrude(p, x)
    ref = c << path
    c.add_ports(ref.ports)
    c.info.length = gf.snap.snap_to_grid(length)
    c.info.width = float(x.info["width"])
    if length > 0 and with_cladding_box and x.info["layers_cladding"]:
        layers_cladding = x.info["layers_cladding"]
        cladding_offset = x.info["cladding_offset"]
        points = get_padding_points(
            component=c,
            default=0,
            bottom=cladding_offset,
            top=cladding_offset,
        )
        for layer in layers_cladding or []:
            c.add_polygon(points, layer=layer)
    c.absorb(ref)
    return c


if __name__ == "__main__":
    # c = straight(cross_section=gf.partial(gf.cross_section.metal3, width=2))

    # c = straight(cross_section=gf.partial(gf.cross_section.strip, width=2))
    # c = straight(cladding_offset=2.5)
    # c = straight(width=2.5)
    c = straight(length=1)
    c.assert_ports_on_grid()
    c.show()
    c.pprint()



from gdsfactory.components.bend_s import bend_s
from gdsfactory.port import Port
from gdsfactory.routing.sort_ports import sort_ports
from gdsfactory.types import Route


def get_route_sbend(port1: Port, port2: Port, **kwargs) -> Route:
    """Returns an Sbend Route to connect two ports.

    Args:
        port1: start port
        port2: end port

    keyword Args:
        nb_points: number of points
        with_cladding_box: square bounding box to avoid DRC errors
        cross_section: function
        kwargs: cross_section settings
    """
    ysize = port2.midpoint[1] - port1.midpoint[1]
    xsize = port2.midpoint[0] - port1.midpoint[0]
    size = (xsize, ysize)

    bend = bend_s(size=size, **kwargs)
    bend_ref = bend.ref()
    bend_ref.connect("o1", port1)
    return Route(
        references=[bend_ref], length=bend.info_child["length"], ports=(port1, port2)
    )


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.Component("test_get_route_sbend")
    pitch = 2.0
    ys_left = [0, 10, 20]
    N = len(ys_left)
    ys_right = [(i - N / 2) * pitch for i in range(N)]

    right_ports = [gf.Port(f"R_{i}", (0, ys_right[i]), 0.5, 180) for i in range(N)]
    left_ports = [gf.Port(f"L_{i}", (-50, ys_left[i]), 0.5, 0) for i in range(N)]
    left_ports.reverse()
    right_ports, left_ports = sort_ports(right_ports, left_ports)

    for p1, p2 in zip(right_ports, left_ports):
        route = get_route_sbend(p1, p2, layer=(2, 0))
        c.add(route.references)

    c.show()



import hashlib
import pathlib
from functools import partial
from pathlib import Path

import pandas as pd

from gdsfactory.config import sparameters_path
from gdsfactory.name import clean_value
from gdsfactory.tech import LAYER, LAYER_STACK
from gdsfactory.types import ComponentOrFactory


def _get_sparameters_path(
    component: ComponentOrFactory,
    dirpath: Path = sparameters_path,
    **kwargs,
) -> Path:
    """Return Sparameters CSV filepath.
    hashes of all simulation settings to get a consitent and unique name.

    Args:
        component: component or component factory.
        dirpath: directory path to store sparameters
        kwargs: simulation settings
    """

    component = component() if callable(component) else component

    dirpath = pathlib.Path(dirpath)
    dirpath = (
        dirpath / component.function_name
        if hasattr(component, "function_name")
        else dirpath
    )
    dirpath.mkdir(exist_ok=True, parents=True)

    kwargs_list = [f"{key}={clean_value(kwargs[key])}" for key in sorted(kwargs.keys())]
    kwargs_string = "_".join(kwargs_list)
    kwargs_hash = hashlib.md5(kwargs_string.encode()).hexdigest()[:8]
    return dirpath / f"{component.name}_{kwargs_hash}.csv"


def _get_sparameters_data(**kwargs) -> pd.DataFrame:
    """Returns Sparameters data in a pandas DataFrame.

    Keyword Args:
        component: component
        dirpath: directory path to store sparameters
        kwargs: simulation settings
    """
    filepath = _get_sparameters_path(**kwargs)
    return pd.read_csv(filepath)


get_sparameters_path_meep = partial(_get_sparameters_path, tool="meep")
get_sparameters_path_lumerical = partial(_get_sparameters_path, tool="lumerical")

get_sparameters_data_meep = partial(_get_sparameters_data, tool="meep")
get_sparameters_data_lumerical = partial(_get_sparameters_data, tool="lumerical")


def test_get_sparameters_path(test: bool = True) -> None:
    import gdsfactory as gf

    layer_to_thickness_sample = {
        LAYER.WG: 220e-3,
        LAYER.SLAB90: 90e-3,
    }
    layer_to_material_sample = {
        LAYER.WG: "si",
        LAYER.SLAB90: "si",
    }

    name1 = "straight_d6c50235"
    name2 = "straight_75fbe695_d6c50235"
    name3 = "straight_75fbe695_6907cda4"
    name4 = "straight_eb75434e"

    c = gf.components.straight()
    p = get_sparameters_path_lumerical(
        component=c,
        layer_to_thickness=layer_to_thickness_sample,
        layer_to_material=layer_to_material_sample,
    )
    if test:
        assert p.stem == name1, p.stem
    else:
        print(f"name1 = {p.stem!r}")

    c = gf.components.straight(layer=LAYER.SLAB90)
    p = get_sparameters_path_lumerical(
        c,
        layer_to_thickness=layer_to_thickness_sample,
        layer_to_material=layer_to_material_sample,
    )
    if test:
        assert p.stem == name2, p.stem
    else:
        print(f"name2 = {p.stem!r}")

    c = gf.components.straight(layer=LAYER.SLAB90)
    p = get_sparameters_path_meep(c, layer_stack=LAYER_STACK)

    if test:
        assert p.stem == name3, p.stem
    else:
        print(f"name3 = {p.stem!r}")

    c = gf.components.straight()
    p = get_sparameters_path_meep(
        component=c,
        layer_to_thickness=layer_to_thickness_sample,
        layer_to_material=layer_to_material_sample,
    )
    if test:
        assert p.stem == name4, p.stem
    else:
        print(f"name4 = {p.stem!r}")


if __name__ == "__main__":
    # import gdsfactory as gf
    # c = gf.components.straight()
    # p = get_sparameters_path(c)
    # print(p)

    test_get_sparameters_path(test=True)



from typing import Optional

import tidy3d as td
from tidy3d.material_library import material_library

MATERIAL_NAME_TO_MEDIUM = {
    "si": material_library["cSi"]["Li1993_293K"],
    "csi": material_library["cSi"]["Li1993_293K"],
    "sio2": material_library["SiO2"]["Horiba"],
    "sin": material_library["Si3N4"]["Horiba"],
}


def get_epsilon(
    wavelength: float = 1.55,
    name: str = "cSi",
) -> float:
    """Return permittivity from material database.

    Args:
        wavelength: wavelength (um)
        name: material name
    """
    medium = get_medium(name=name)
    frequency = td.C_0 / wavelength
    return medium.eps_model(frequency)


def get_medium(name: str, n: Optional[float] = None) -> td.Medium:
    """Return Medium from materials database

    Args:
        name: material name
        n: optional index

    """

    name = name.lower()

    if n:
        m = td.Medium(permittivity=n ** 2)

    elif name == "sio2":
        m = td.Medium(permittivity=1.45 ** 2)
    elif name in ["csi", "si"]:
        m = td.Medium(permittivity=3.48 ** 2)

    elif name in ["sin", "si3n4"]:
        m = td.Medium(permittivity=2.0 ** 2)
    elif name in MATERIAL_NAME_TO_MEDIUM:
        m = MATERIAL_NAME_TO_MEDIUM[name]

    else:
        materials = list(MATERIAL_NAME_TO_MEDIUM.keys())

        raise ValueError(f"Material {name!r} not in {materials}")

    return m


if __name__ == "__main__":
    print(get_epsilon(name="cSi") ** 0.5)
    # m = get_medium(name="SiO2")
    # m = td.Medium(permittivity=1.45 ** 2)



from gdsfactory.add_grating_couplers import (
    add_grating_couplers,
    add_grating_couplers_with_loopback_fiber_single,
)
from gdsfactory.add_padding import add_padding_container
from gdsfactory.add_termination import add_termination
from gdsfactory.components.bend_port import bend_port
from gdsfactory.components.cavity import cavity
from gdsfactory.components.extension import extend_ports
from gdsfactory.components.ring_single_dut import ring_single_dut
from gdsfactory.functions import rotate
from gdsfactory.routing.add_electrical_pads_shortest import add_electrical_pads_shortest
from gdsfactory.routing.add_electrical_pads_top import add_electrical_pads_top
from gdsfactory.routing.add_fiber_array import add_fiber_array
from gdsfactory.routing.add_fiber_single import add_fiber_single
from gdsfactory.routing.fanout2x2 import fanout2x2

container_factory = dict(
    add_electrical_pads_shortest=add_electrical_pads_shortest,
    add_electrical_pads_top=add_electrical_pads_top,
    add_fiber_array=add_fiber_array,
    add_fiber_single=add_fiber_single,
    add_grating_couplers=add_grating_couplers,
    add_grating_couplers_with_loopback_fiber_single=add_grating_couplers_with_loopback_fiber_single,
    # add_grating_couplers_with_loopback_fiber_array=add_grating_couplers_with_loopback_fiber_array,
    add_padding_container=add_padding_container,
    add_termination=add_termination,
    cavity=cavity,
    extend_ports=extend_ports,
    fanout2x2=fanout2x2,
    ring_single_dut=ring_single_dut,
    bend_port=bend_port,
    rotate=rotate,
)


__all__ = list(container_factory.keys())



import pydantic
import pytest

import gdsfactory as gf


@gf.cell
def component_with_straight(component: gf.Component) -> gf.Component:
    c = gf.Component()
    c.add_ref(component)
    c.add_ref(gf.components.straight())
    return c


def test_validator_pass():
    component = gf.components.straight(length=10)
    component_with_straight(component=component)


# def test_validator_fail_empty():
#     component = gf.Component()
#     with pytest.raises(pydantic.ValidationError):
#         component_with_straight(component=component)


def test_validator_fail_name_too_long():
    component = gf.Component(name="a" * 200)

    # component_with_straight(component=component)
    with pytest.raises(pydantic.ValidationError):
        component_with_straight(component=component)


if __name__ == "__main__":
    # test_validator_pass()
    # test_validator_fail_empty()
    test_validator_fail_name_too_long()



import gdsfactory as gf
from gdsfactory.components.mzi import mzi
from gdsfactory.components.straight_heater_metal import straight_heater_metal

mzi_phase_shifter = gf.partial(mzi, length_x=None)

mzi_phase_shifter_top_heater_metal = gf.partial(
    mzi_phase_shifter, straight_x_top=straight_heater_metal
)

if __name__ == "__main__":
    # c = mzi_phase_shifter_top_heater_metal()
    # c = mzi_phase_shifter(
    #     straight_x_top=gf.components.straight_pin, straight_x_bot=gf.components.straight_pin
    # )
    # c = mzi_phase_shifter(
    #     # straight_x_top=gf.components.straight_heater_doped_rib,
    #     straight_x_bot=gf.components.straight_heater_doped_rib,
    #     delta_length=20,
    #     length_x=500,
    # )
    c = mzi_phase_shifter()
    c.show()
    print(c.name)



from gdsfactory.routing.get_bundle import get_bundle, get_bundle_same_axis_no_grouping
from gdsfactory.routing.get_bundle_from_waypoints import get_bundle_from_waypoints
from gdsfactory.routing.get_bundle_path_length_match import get_bundle_path_length_match

routing_strategy = dict(
    get_bundle=get_bundle,
    get_bundle_path_length_match=get_bundle_path_length_match,
    get_bundle_same_axis_no_grouping=get_bundle_same_axis_no_grouping,
    get_bundle_from_waypoints=get_bundle_from_waypoints,
)



from typing import Dict, List, Optional, Tuple

from pydantic import BaseModel

Layer = Tuple[int, int]
Point = Tuple[float, float]
Points = List[Point]


class SectionModel(BaseModel):
    """
    Args:
        width: of the section
        offset: center to center
        layer:
        ports: optional name of the ports
        name: optional section name
        port_types:
    """

    width: float
    offset: float = 0
    layer: Layer = (1, 0)
    ports: Tuple[Optional[str], Optional[str]] = (None, None)
    name: Optional[str] = None
    port_types: Tuple[str, str] = ("optical", "optical")


class CrossSectionModel(BaseModel):
    width: float = 0.5
    layer: Layer = (1, 0)
    width_wide: Optional[float] = None
    auto_widen: bool = False
    auto_widen_minimum_length: float = 200.0
    taper_length: float = 10.0
    radius: float = 10.0
    cladding_offset: float = 3.0
    layers_cladding: Optional[Tuple[Layer, ...]] = None
    sections: Optional[Tuple[SectionModel, ...]] = None
    port_names: Tuple[str, str] = ("o1", "o2")
    port_types: Tuple[str, str] = ("optical", "optical")
    min_length: float = 10e-3
    start_straight_length: float = 10e-3
    end_straight_length: float = 10e-3
    snap_to_grid: Optional[float] = None


class PathModel(BaseModel):
    points: Points


class PathStraightModel(BaseModel):
    length: float
    npoints: int = 2


class PathEulerModel(BaseModel):
    radius: float = 10
    angle: int = 90
    p: float = 0.5
    use_eff: bool = False
    npoints: int = 720


class PathArcModel(BaseModel):
    radius: float = 10
    angle: int = 90
    npoints: int = 720


class TechModel(BaseModel):
    cross_sections: Dict[str, CrossSectionModel]
    layers: Dict[str, Layer]


class ComponentTwoPortsModel(BaseModel):
    cross_section: CrossSectionModel
    path: PathModel


if __name__ == "__main__":
    x = CrossSectionModel()
    t = TechModel.schema_json()



import dataclasses
from typing import Callable, Dict, List, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
from meep import mpb
from pydantic import BaseModel
from scipy.interpolate import RectBivariateSpline

# cmap_default = 'viridis'
cmap_default = "RdBu"


@dataclasses.dataclass
class Mode:
    mode_number: int
    wavelength: float
    neff: float
    ng: Optional[float] = None
    fraction_te: Optional[float] = None
    fraction_tm: Optional[float] = None
    effective_area: Optional[float] = None
    E: Optional[np.ndarray] = None
    H: Optional[np.ndarray] = None
    eps: Optional[np.ndarray] = None
    y: Optional[np.ndarray] = None
    z: Optional[np.ndarray] = None

    def __repr__(self):
        return f"Mode{self.mode_number}"

    def grid_interp(self):
        """
        Creates new attributes with scipy.interpolate.RectBivariateSpline objects that can be used to interpolate the field on a new regular grid

        Args:
            y_grid (np.array): y values where to evaluate, in increasing array
            z_grid (np.array): z values where to evaluate, in increasing array
        """

        self.Ex_grid_interp = lambda y_arr, z_arr: np.flip(
            RectBivariateSpline(self.y, self.z, np.real(self.E[:, :, 0, 0]))(
                y_arr, z_arr, grid=True
            )
            + (
                1j
                * RectBivariateSpline(self.y, self.z, np.imag(self.E[:, :, 0, 0]))(
                    y_arr, z_arr, grid=True
                )
            )
        )
        self.Ey_grid_interp = lambda y_arr, z_arr: np.flip(
            RectBivariateSpline(self.y, self.z, np.real(self.E[:, :, 0, 1]))(
                y_arr, z_arr, grid=True
            )
            + (
                1j
                * RectBivariateSpline(self.y, self.z, np.imag(self.E[:, :, 0, 1]))(
                    y_arr, z_arr, grid=True
                )
            )
        )
        self.Ez_grid_interp = lambda y_arr, z_arr: np.flip(
            RectBivariateSpline(self.y, self.z, np.real(self.E[:, :, 0, 2]))(
                y_arr, z_arr, grid=True
            )
            + (
                1j
                * RectBivariateSpline(self.y, self.z, np.imag(self.E[:, :, 0, 2]))(
                    y_arr, z_arr, grid=True
                )
            )
        )
        self.Hx_grid_interp = lambda y_arr, z_arr: np.flip(
            RectBivariateSpline(self.y, self.z, np.real(self.H[:, :, 0, 0]))(
                y_arr, z_arr, grid=True
            )
            + (
                1j
                * RectBivariateSpline(self.y, self.z, np.imag(self.H[:, :, 0, 0]))(
                    y_arr, z_arr, grid=True
                )
            )
        )
        self.Hy_grid_interp = lambda y_arr, z_arr: np.flip(
            RectBivariateSpline(self.y, self.z, np.real(self.H[:, :, 0, 1]))(
                y_arr, z_arr, grid=True
            )
            + (
                1j
                * RectBivariateSpline(self.y, self.z, np.imag(self.H[:, :, 0, 1]))(
                    y_arr, z_arr, grid=True
                )
            )
        )
        self.Hz_grid_interp = lambda y_arr, z_arr: np.flip(
            RectBivariateSpline(self.y, self.z, np.real(self.H[:, :, 0, 2]))(
                y_arr, z_arr, grid=True
            )
            + (
                1j
                * RectBivariateSpline(self.y, self.z, np.imag(self.H[:, :, 0, 2]))(
                    y_arr, z_arr, grid=True
                )
            )
        )
        return 1

    def plot_eps(
        self,
        cmap: str = "binary",
        origin="lower",
        logscale: bool = False,
        show: bool = True,
    ):
        """plot index profle"""
        plt.imshow(
            self.eps ** 0.5,
            cmap=cmap,
            origin=origin,
            aspect="auto",
            extent=[np.min(self.y), np.max(self.y), np.min(self.z), np.max(self.z)],
        )
        plt.title("index profile")
        plt.ylabel("z-axis")
        plt.xlabel("y-axis")
        plt.colorbar()
        if show:
            plt.show()

    def plot_e(
        self,
        cmap: str = cmap_default,
        origin="lower",
        logscale: bool = False,
        show: bool = True,
        scale: bool = False,
    ):
        """ """
        E = self.E / abs(max(self.E.min(), self.E.max(), key=abs)) if scale else self.E
        Eabs = np.sqrt(
            np.multiply(E[:, :, 0, 2], E[:, :, 0, 2])
            + np.multiply(E[:, :, 0, 1], E[:, :, 0, 1])
            + np.multiply(E[:, :, 0, 0], E[:, :, 0, 0])
        )
        ep = abs(Eabs)
        ep = 10 * np.log10(ep) if logscale else ep
        plt.imshow(
            ep.T,
            cmap=cmap,
            origin=origin,
            aspect="auto",
            extent=[np.min(self.y), np.max(self.y), np.min(self.z), np.max(self.z)],
            vmin=0 if scale else ep.min(),
            vmax=1 if scale else ep.max(),
        )
        plt.title("$|E|$")
        plt.ylabel("z-axis")
        plt.xlabel("y-axis")
        plt.colorbar()
        if show:
            plt.show()

    def plot_ex(
        self,
        cmap: str = cmap_default,
        origin="lower",
        logscale: bool = False,
        show: bool = True,
        scale: bool = False,
        operation: Callable = np.real,
    ):
        E = self.E / abs(max(self.E.min(), self.E.max(), key=abs)) if scale else self.E
        ex = E[:, :, 0, 0]
        ex = 10 * np.log10(np.abs(ex)) if logscale else operation(ex)
        plt.imshow(
            ex.T,
            cmap=cmap,
            origin=origin,
            aspect="auto",
            extent=[np.min(self.y), np.max(self.y), np.min(self.z), np.max(self.z)],
            vmin=-1 if scale else ex.min(),
            vmax=1 if scale else ex.max(),
        )
        plt.title("{}($E_x$)".format(operation))
        plt.ylabel("z-axis")
        plt.xlabel("y-axis")
        plt.colorbar()
        if show:
            plt.show()

    def plot_ey(
        self,
        cmap: str = cmap_default,
        origin="lower",
        logscale: bool = False,
        show: bool = True,
        scale: bool = False,
        operation: Callable = np.real,
    ):
        E = self.E / abs(max(self.E.min(), self.E.max(), key=abs)) if scale else self.E
        ey = E[:, :, 0, 1]
        ey = 10 * np.log10(np.abs(ey)) if logscale else operation(ey)
        plt.imshow(
            ey.T,
            cmap=cmap,
            origin=origin,
            aspect="auto",
            extent=[np.min(self.y), np.max(self.y), np.min(self.z), np.max(self.z)],
            vmin=-1 if scale else ey.min(),
            vmax=1 if scale else ey.max(),
        )
        plt.title("{}($E_y$)".format(operation))
        plt.ylabel("z-axis")
        plt.xlabel("y-axis")
        plt.colorbar()
        if show:
            plt.show()

    def plot_ez(
        self,
        cmap: str = cmap_default,
        origin="lower",
        logscale: bool = False,
        show: bool = True,
        scale: bool = False,
        operation: Callable = np.real,
    ):
        E = self.E / abs(max(self.E.min(), self.E.max(), key=abs)) if scale else self.E
        ez = E[:, :, 0, 2]
        ez = 10 * np.log10(ez) if logscale else operation(ez)
        plt.imshow(
            ez.T,
            cmap=cmap,
            origin=origin,
            aspect="auto",
            extent=[np.min(self.y), np.max(self.y), np.min(self.z), np.max(self.z)],
            vmin=-1 if scale else ez.min(),
            vmax=1 if scale else ez.max(),
        )
        plt.title("{}($E_z$)".format(operation))
        plt.ylabel("z-axis")
        plt.xlabel("y-axis")
        plt.colorbar()
        if show:
            plt.show()

    def plot_e_all(
        self,
        cmap: str = cmap_default,
        origin="lower",
        logscale: bool = False,
        show: bool = True,
        scale: bool = False,
        operation: Callable = np.real,
    ):
        plt.figure(figsize=(16, 10), dpi=100)

        plt.subplot(2, 3, 1)
        self.plot_ex(show=False, scale=scale, cmap=cmap, operation=operation)

        plt.subplot(2, 3, 2)
        self.plot_ey(show=False, scale=scale, cmap=cmap, operation=operation)

        plt.subplot(2, 3, 3)
        self.plot_ez(show=False, scale=scale, cmap=cmap, operation=operation)

        plt.subplot(2, 3, 4)
        self.plot_e(show=False, scale=scale)

        plt.subplot(2, 3, 5)
        self.plot_eps(show=False)

        plt.tight_layout()
        plt.show()

    def plot_h(
        self,
        cmap: str = cmap_default,
        origin="lower",
        logscale: bool = False,
        show: bool = True,
        scale: bool = False,
    ):
        H = self.H / abs(max(self.H.min(), self.H.max(), key=abs)) if scale else self.H
        Habs = np.sqrt(
            np.multiply(H[:, :, 0, 2], H[:, :, 0, 2])
            + np.multiply(H[:, :, 0, 1], H[:, :, 0, 1])
            + np.multiply(H[:, :, 0, 0], H[:, :, 0, 0])
        )
        hp = abs(Habs)
        hp = 10 * np.log10(hp) if logscale else hp
        plt.imshow(
            hp.T,
            cmap=cmap,
            origin=origin,
            aspect="auto",
            extent=[np.min(self.y), np.max(self.y), np.min(self.z), np.max(self.z)],
            vmin=0 if scale else hp.min(),
            vmax=1 if scale else hp.max(),
        )
        plt.title("$|H|$")
        plt.ylabel("z-axis")
        plt.xlabel("y-axis")
        plt.colorbar()
        if show:
            plt.show()

    def plot_hx(
        self,
        cmap: str = cmap_default,
        origin="lower",
        logscale: bool = False,
        show: bool = True,
        scale: bool = False,
        operation: Callable = np.real,
    ):
        H = self.H / abs(max(self.H.min(), self.H.max(), key=abs)) if scale else self.H
        hx = H[:, :, 0, 0]
        hx = 10 * np.log10(np.abs(hx)) if logscale else operation(hx)
        plt.imshow(
            hx.T,
            cmap=cmap,
            origin=origin,
            aspect="auto",
            extent=[np.min(self.y), np.max(self.y), np.min(self.z), np.max(self.z)],
            vmin=-1 if scale else hx.min(),
            vmax=1 if scale else hx.max(),
        )
        plt.title("{}($H_x$)".format(operation))
        plt.ylabel("z-axis")
        plt.xlabel("y-axis")
        plt.colorbar()
        if show:
            plt.show()

    def plot_hy(
        self,
        cmap: str = cmap_default,
        origin="lower",
        logscale: bool = False,
        show: bool = True,
        scale: bool = False,
        operation: Callable = np.real,
    ):
        H = self.H / abs(max(self.H.min(), self.H.max(), key=abs)) if scale else self.H
        hy = H[:, :, 0, 1]
        hy = 10 * np.log10(np.abs(hy)) if logscale else operation(hy)
        plt.imshow(
            hy.T,
            cmap=cmap,
            origin=origin,
            aspect="auto",
            extent=[np.min(self.y), np.max(self.y), np.min(self.z), np.max(self.z)],
            vmin=-1 if scale else hy.min(),
            vmax=1 if scale else hy.max(),
        )
        plt.title("{}($H_y$)".format(operation))
        plt.ylabel("z-axis")
        plt.xlabel("y-axis")
        plt.colorbar()
        if show:
            plt.show()

    def plot_hz(
        self,
        cmap: str = cmap_default,
        origin="lower",
        logscale: bool = False,
        show: bool = True,
        scale: bool = False,
        operation: Callable = np.real,
    ):
        H = self.H / abs(max(self.H.min(), self.H.max(), key=abs)) if scale else self.H
        hz = abs(H[:, :, 0, 2])
        hz = 10 * np.log10(hz) if logscale else operation(hz)
        plt.imshow(
            hz.T,
            cmap=cmap,
            origin=origin,
            aspect="auto",
            extent=[np.min(self.y), np.max(self.y), np.min(self.z), np.max(self.z)],
            vmin=0 if scale else hz.min(),
            vmax=1 if scale else hz.max(),
        )
        plt.title("{}($H_z$)".format(operation))
        plt.ylabel("z-axis")
        plt.xlabel("y-axis")
        plt.colorbar()
        if show:
            plt.show()

    def plot_h_all(
        self,
        cmap: str = cmap_default,
        origin="lower",
        logscale: bool = False,
        show: bool = True,
        scale: bool = False,
        operation: Callable = np.real,
    ):
        plt.figure(figsize=(16, 10), dpi=100)

        plt.subplot(2, 3, 1)
        self.plot_hx(show=False, scale=scale, cmap=cmap, operation=operation)

        plt.subplot(2, 3, 2)
        self.plot_hy(show=False, scale=scale, cmap=cmap, operation=operation)

        plt.subplot(2, 3, 3)
        self.plot_hz(show=False, scale=scale, cmap=cmap, operation=operation)

        plt.subplot(2, 3, 4)
        self.plot_h(show=False, scale=scale)

        plt.subplot(2, 3, 5)
        self.plot_eps(show=False)

        plt.tight_layout()
        plt.show()


@dataclasses.dataclass
class Waveguide:
    """
    Args:
        wg_width: float
        wg_thickness: float
        slab_thickness: float
        ncore: float = 3.47
        nclad: float = 1.44
        sy: float
        sz: float
        resolution: int
        nmodes: int
        modes: Dict[Mode]

    """

    wg_width: float = 0.45
    wg_thickness: float = 0.22
    slab_thickness: float = 0
    ncore: float = 3.47
    nclad: float = 1.44
    sy: float = 2.0
    sz: float = 2.0
    resolution: int = 32
    nmodes: int = 4
    modes: Optional[Dict[int, Mode]] = None


class WavelengthSweep(BaseModel):
    wavelength: List[float]
    neff: Dict[int, List[float]]
    ng: Dict[int, List[float]]


ModeSolverFactory = Callable[..., mpb.ModeSolver]
ModeSolverOrFactory = Union[mpb.ModeSolver, ModeSolverFactory]


if __name__ == "__main__":
    import gdsfactory.simulation.modes as gm

    m = gm.find_modes()
    # m[1].plot_e_all(operation=np.abs)
    # plt.show()
    # m[1].plot_h_all(operation=np.abs)
    # w = Waveguide()

    ys = np.linspace(-3, 3, 2000)
    zs = np.linspace(-1.5, 1.5, 2000)
    m[1].grid_interp()

    Ex_interp = m[1].Ex_grid_interp(ys, zs)
    Ey_interp = m[1].Ey_grid_interp(ys, zs)
    Ez_interp = m[1].Ez_grid_interp(ys, zs)

    plt.figure(figsize=(10, 8), dpi=100)

    plt.subplot(3, 2, 1)
    plt.title("As calculated")
    m[1].plot_ex(show=False, operation=np.abs, scale=False)

    plt.subplot(3, 2, 2)
    plt.title("Interp")
    plt.imshow(
        np.abs(Ex_interp).T,
        aspect="auto",
        extent=[np.min(ys), np.max(ys), np.min(zs), np.max(zs)],
    )
    plt.colorbar()

    plt.subplot(3, 2, 3)
    plt.title("As calculated")
    m[1].plot_ey(show=False, operation=np.abs, scale=False)

    plt.subplot(3, 2, 4)
    plt.title("Interp")
    plt.imshow(
        np.abs(Ey_interp).T,
        aspect="auto",
        extent=[np.min(ys), np.max(ys), np.min(zs), np.max(zs)],
    )
    plt.colorbar()

    plt.subplot(3, 2, 5)
    plt.title("As calculated")
    m[1].plot_ez(show=False, operation=np.abs, scale=False)

    plt.subplot(3, 2, 6)
    plt.title("Interp")
    plt.imshow(
        np.abs(Ez_interp).T,
        aspect="auto",
        extent=[np.min(ys), np.max(ys), np.min(zs), np.max(zs)],
    )
    plt.colorbar()

    plt.show()



""" Lets create a new component.

We create a function which returns a gf.Component.

Lets build straight crossing out of a vertical and horizonal arm

- Create a component using component_factory function. You should the cell decorator to define the name automatically and uniquely.
- Define the polygons in the component
- Add ports to the component so you can connect it with other components
"""

from phidl import quickplot as plot

import gdsfactory as gf
from gdsfactory import LAYER
from gdsfactory.component import Component


@gf.cell
def test_crossing_arm(
    wg_width: float = 0.5,
    r1: float = 3.0,
    r2: float = 1.1,
    taper_width: float = 1.2,
    taper_length: float = 3.4,
) -> Component:
    """crossing arm"""
    c = gf.Component()
    c << gf.components.ellipse(radii=(r1, r2), layer=LAYER.SLAB150)

    xmax = taper_length + taper_width / 2
    h = wg_width / 2
    taper_points = [
        (-xmax, h),
        (-taper_width / 2, taper_width / 2),
        (taper_width / 2, taper_width / 2),
        (xmax, h),
        (xmax, -h),
        (taper_width / 2, -taper_width / 2),
        (-taper_width / 2, -taper_width / 2),
        (-xmax, -h),
    ]

    c.add_polygon(taper_points, layer=LAYER.WG)

    c.add_port(
        name="o1", midpoint=(-xmax, 0), orientation=180, width=wg_width, layer=LAYER.WG
    )

    c.add_port(
        name="o2", midpoint=(xmax, 0), orientation=0, width=wg_width, layer=LAYER.WG
    )
    return c


if __name__ == "__main__":
    c = test_crossing_arm()
    c.show()
    plot(c)



"""Write DRC rule decks in klayout.

TODO:

- add min area
- define derived layers (composed rules)
"""

import pathlib
from dataclasses import asdict, is_dataclass
from typing import List, Optional

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from gdsfactory.config import logger
from gdsfactory.install import get_klayout_path
from gdsfactory.types import Dict, Layer, PathType

layer_name_to_min_width: Dict[str, float]

RuleType = Literal[
    "width",
    "space",
    "enclosing",
]


def rule_width(value: float, layer: str, angle_limit: float = 90) -> str:
    """Min feature size"""
    category = "width"
    error = f"{layer} {category} {value}um"
    return (
        f"{layer}.{category}({value}, angle_limit({angle_limit}))"
        f".output('{error}', '{error}')"
    )


def rule_space(value: float, layer: str, angle_limit: float = 90) -> str:
    """Min Space between shapes of layer"""
    category = "space"
    error = f"{layer} {category} {value}um"
    return (
        f"{layer}.{category}({value}, angle_limit({angle_limit}))"
        f".output('{error}', '{error}')"
    )


def rule_separation(value: float, layer1: str, layer2: str):
    """Min space between different layers"""
    error = f"min {layer1} {layer2} separation {value}um"
    return f"{layer1}.separation({layer2}, {value})" f".output('{error}', '{error}')"


def rule_enclosing(
    value: float, layer1: str, layer2: str, angle_limit: float = 90
) -> str:
    """Layer1 must be enclosed by layer2 by value.
    checks if layer1 encloses (is bigger than) layer2 by value
    """
    error = f"{layer1} enclosing {layer2} by {value}um"
    return (
        f"{layer1}.enclosing({layer2}, angle_limit({angle_limit}), {value})"
        f".output('{error}', '{error}')"
    )


def write_layer_definition(layer_map: Dict[str, Layer]) -> str:
    """Returns layer_map definition script for klayout

    Args:
        layer_map: can be dict or dataclass

    """
    layer_map = asdict(layer_map) if is_dataclass(layer_map) else layer_map
    return [
        f"{key} = input({value[0]}, {value[1]})" for key, value in layer_map.items()
    ]


def write_drc_deck(rules: List[str], layer_map: Dict[str, Layer]) -> str:
    """Returns drc_rule_deck for klayou

    Args:
        rules: list of rules
        layer_map: layer definitions can be dict or dataclass

    """
    script = []
    script += write_layer_definition(layer_map=layer_map)
    script += ["\n"]
    script += rules
    return "\n".join(script)


def write_drc_deck_macro(
    name="generic",
    filepath: Optional[PathType] = None,
    shortcut: str = "Ctrl+Shift+D",
    **kwargs,
) -> str:
    """Write script for klayout rule deck

    Args:
        name: drc rule deck name
        filepath: Optional macro path (defaults to .klayout/drc/name.lydrc)

    Keyword Args:
        rules: list of rules
        layer_map: layer definitions can be dict or dataclass

    Keyword Args:
        rules: list of rules
        layer_map: layer definitions can be dict or dataclass

    """
    script = f"""<?xml version="1.0" encoding="utf-8"?>
<klayout-macro>
 <description>{name} DRC</description>
 <version/>
 <category>drc</category>
 <prolog/>
 <epilog/>
 <doc/>
 <autorun>false</autorun>
 <autorun-early>false</autorun-early>
 <shortcut>{shortcut}</shortcut>
 <show-in-menu>true</show-in-menu>
 <group-name>drc_scripts</group-name>
 <menu-path>tools_menu.drc.end</menu-path>
 <interpreter>dsl</interpreter>
 <dsl-interpreter-name>drc-dsl-xml</dsl-interpreter-name>
 <text># {name} DRC

# Read about DRC scripts in the User Manual under "Design Rule Check (DRC)"
# Based on SOEN pdk https://github.com/usnistgov/SOEN-PDK/tree/master/tech/OLMAC
# http://klayout.de/doc/manual/drc_basic.html

report("generic DRC")
tiles(100)
tile_borders(2)
threads(3)
"""
    script += write_drc_deck(**kwargs)

    script += """
</text>
</klayout-macro>
"""
    filepath = filepath or get_klayout_path() / "drc" / f"{name}.lydrc"
    filepath = pathlib.Path(filepath)
    filepath.write_text(script)
    logger.info(f"Wrote DRC deck to {filepath}")
    return script


if __name__ == "__main__":
    import gdsfactory as gf

    rules = [
        rule_width(layer="WG", value=0.2),
        rule_space(layer="WG", value=0.2),
        rule_width(layer="M1", value=1),
        rule_width(layer="M2", value=2),
        rule_space(layer="M2", value=2),
        rule_separation(layer1="HEATER", layer2="M1", value=1.0),
        rule_enclosing(layer1="M1", layer2="VIAC", value=0.2),
    ]

    drc_rule_deck = write_drc_deck_macro(rules=rules, layer_map=gf.LAYER)
    print(drc_rule_deck)



import glob
import json
import re
from collections import defaultdict

import klayout.db as pya

from gdsfactory.autoplacer.cell_list import CellList
from gdsfactory.autoplacer.functions import WORKING_MEMORY, area


class Library(object):
    """Library of cells with convenient methods to:

        - load from disk
        - pop groups
        - add padding

    Args:
        root: GDS devices path

    To make a `Library` containing all the devices in `build/devices`, just instantiate the class (`library = Library()`).
    You can then pull out subsets of devices using `library.pop()`.
    `library.pop()` accepts a regular expression, which is matched against the top-cell name, to select the desired cells.
    It also has `padding` and `normalize` arguments.

    By default, padding is added to cells when they are popped from the `Library`.
    Attempts are also made to normalize the form-factor of the devices so that the grating couplers are aligned.
    This will come at the cost of some space.
    If you are packing many disparate components, you should either group them into blocks first, or pass `normalize=False`.

    .. code::

        import gdsfactory.autoplacer as ap

        # Load all the GDS files from build/devices
        library = ap.Library()

        # Select all devices with "ring" and "euler" in the top-cell name.
        library.pop("ring.*euler.*")

        # Same selection, less padding
        library.pop("ring.*euler.*", padding=0)

        # Same selection, without attempting to align gratings
        library.pop("ring.*euler.*", normalize=False)

    """

    def __init__(self, root: str = "build/devices") -> None:
        self.root = root
        self.cells = {}
        self.does = defaultdict(list)
        self.load_all_gds()
        self.load_all_json()

    def load_all_gds(self) -> None:
        """Loads all the gds files"""
        filenames = glob.glob(self.root + "/*.gds")
        print("Loading {} GDS files...".format(len(filenames)))
        for filename in filenames:
            self.load_gds(filename)
        print("Done")

    def load_all_json(self) -> None:
        """loads all the json files"""
        filenames = glob.glob(self.root + "/*.json")
        for filename in filenames:
            self.load_json(filename)

    def load_gds(self, filename):
        """Load a GDS and append it into self.cells"""
        layout = pya.Layout()
        WORKING_MEMORY[filename] = layout
        layout.read(str(filename))
        self.cells[layout.top_cell().name] = layout.top_cell()
        self.cells[layout.top_cell().name].metadata = {}

    def load_json(self, filename):
        """Load json metadata"""
        with open(filename) as f:
            metadata = json.load(f)
            if metadata.get("type") == "sweep":
                doe_name = metadata.get("name")
                for cell_name in metadata.get("cells"):
                    if cell_name in self.cells:
                        self.does[doe_name].append(self.cells[cell_name])

    def get(self, regex):
        cells = [
            cell
            for key, cell in self.cells.items()
            if re.search(regex, key, flags=re.IGNORECASE)
        ]
        return CellList(cells)

    def pop_doe(self, regex):
        """pop out a set of cells"""
        cells = []
        if regex in self.does:
            cells = self.does[regex]
            del self.does[regex]
            self.delete_cells(cells)
            cells = sorted(cells, key=area, reverse=True)

        else:
            print("Warning: no cells found for {}".format(regex))

        return CellList(cells)

    def pop(self, regex: str, delete: bool = True) -> CellList:
        """pop cells"""
        # pop out the cells
        # cells = [
        #     cell for key, cell in self.cells.items()
        #     if re.search(regex, key, flags=re.IGNORECASE)
        # ]
        cells = []
        keys = []
        for key, cell in self.cells.items():
            if re.search(regex, key, flags=re.IGNORECASE):
                cells.append(cell)
                keys.append(key)

        if delete:
            for key in keys:
                del self.cells[key]
        if cells:
            cells = sorted(cells, key=area, reverse=True)
        else:
            print("Warning: no cells found for {}".format(regex))

        return CellList(cells)

    def delete_cells(self, cells):
        for cell in cells:
            try:
                del self.cells[cell.name]
            except KeyError:
                pass

    def list(self):
        """just list the devices currently in the collection"""
        print("Library contains cells:")
        for name in sorted(self.cells.keys()):
            print("-", name)

        if self.does and False:
            print()
            print("DOEs:")
            for key, value in self.does.items():
                print(" - {} ({})".format(key, len(value)))

    def count(self):
        """Safety check at the end"""
        if self.cells:
            print("{} cells were not used".format(len(self.cells)))

    def __str__(self):
        return "<collection of {} cells>".format(len(self.cells))


if __name__ == "__main__":
    lib = Library()



"""Add reference for a grating coupler array."""
from typing import List

import gdsfactory as gf
from gdsfactory.component import ComponentReference
from gdsfactory.port import Port
from gdsfactory.routing.manhattan import round_corners
from gdsfactory.types import ComponentFactory


def gen_loopback(
    start_port: Port,
    end_port: Port,
    gc: ComponentFactory,
    grating_separation: float = 127.0,
    gc_rotation: int = -90,
    gc_port_name: str = "o1",
    bend_radius_loopback: float = 10.0,
    bend: ComponentFactory = gf.components.bend_euler,
    straight: ComponentFactory = gf.components.straight,
    y_bot_align_route: None = None,
) -> List[ComponentReference]:
    """Returns loopback (grating coupler align reference) references
    from a start_port and end_port

    Input grating generated on the left of start_port
    Output grating generated on the right of end_port

    .. code::

        __________________________________________
        | separation  |            |              |
        |             |            |              |
       GC          start_port  end_port          GC

    """

    gc = gc() if callable(gc) else gc

    if hasattr(start_port, "y"):
        y0 = start_port.y
    else:
        y0 = start_port[1]

    if hasattr(start_port, "x"):
        x0 = start_port.x - grating_separation
    else:
        x0 = start_port[0] - grating_separation

    if hasattr(end_port, "x"):
        x1 = end_port.x + grating_separation
    else:
        x1 = end_port[0] + grating_separation

    gca1, gca2 = [
        gc.ref(position=(x, y0), rotation=gc_rotation, port_id=gc_port_name)
        for x in [x0, x1]
    ]

    gsi = gc.size_info
    p0 = gca1.ports[gc_port_name].position
    p1 = gca2.ports[gc_port_name].position
    bend90 = bend(radius=bend_radius_loopback)

    if hasattr(bend90, "dx"):
        a = abs(bend90.info.dy)
    else:
        a = bend_radius_loopback + 0.5
    b = max(2 * a, grating_separation / 2)
    y_bot_align_route = (
        y_bot_align_route if y_bot_align_route is not None else -gsi.width - 5.0
    )

    points = [
        p0,
        p0 + (0, a),
        p0 + (b, a),
        p0 + (b, y_bot_align_route),
        p1 + (-b, y_bot_align_route),
        p1 + (-b, a),
        p1 + (0, a),
        p1,
    ]
    route = round_corners(points=points, bend=bend90, straight=straight)
    elements = [gca1, gca2]
    elements.extend(route.references)
    return elements


@gf.cell
def straight_with_loopback() -> gf.Component:
    c = gf.Component("straight_with_loopback")
    wg = c << gf.components.straight()
    c.add(
        gen_loopback(
            wg.ports["o1"], wg.ports["o2"], gc=gf.components.grating_coupler_te
        )
    )
    return c


if __name__ == "__main__":
    component = straight_with_loopback()
    component.show()



import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.pad import pad as pad_function
from gdsfactory.port import select_ports_electrical
from gdsfactory.routing.route_quad import route_quad
from gdsfactory.types import ComponentOrFactory


@gf.cell
def add_electrical_pads_shortest(
    component: Component,
    pad: ComponentOrFactory = pad_function,
    pad_port_spacing: float = 50.0,
    select_ports=select_ports_electrical,
    port_orientation: int = 90,
    layer: gf.types.Layer = (31, 0),
    **kwargs,
) -> Component:
    """Add pad to each closest electrical port.

    Args:
        component:
        pad: pad element or function
        pad_port_spacing: between pad and port
        select_ports: function
        port_orientation
        layer: for the routing
        **kwargs: pad_settings

    """
    c = Component()
    c.component = component
    ref = c << component
    ports = select_ports(ref.ports)
    ports = list(ports.values())

    pad = pad(**kwargs) if callable(pad) else pad
    pad_port_spacing += pad.info_child.full["size"][0] / 2

    for port in ports:
        p = c << pad
        if port_orientation == 0:
            p.x = port.x + pad_port_spacing
            p.y = port.y
            c.add(route_quad(port, p.ports["e1"], layer=layer))
        elif port_orientation == 180:
            p.x = port.x - pad_port_spacing
            p.y = port.y
            c.add(route_quad(port, p.ports["e3"], layer=layer))
        elif port_orientation == 90:
            p.y = port.y + pad_port_spacing
            p.x = port.x
            c.add(route_quad(port, p.ports["e4"], layer=layer))
        elif port_orientation == 270:
            p.y = port.y - pad_port_spacing
            p.x = port.x
            c.add(route_quad(port, p.ports["e2"], layer=layer))

    c.add_ports(ref.ports)
    for port in ports:
        c.ports.pop(port.name)
    c.copy_child_info(component)
    return c


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.components.cross(length=100, layer=gf.LAYER.M3)
    c = gf.components.mzi_phase_shifter()
    c = gf.components.straight_heater_metal()
    cc = add_electrical_pads_shortest(component=c)
    cc.show()



import numpy as np
from pytest_regressions.data_regression import DataRegressionFixture

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.port import Port
from gdsfactory.routing.get_bundle_from_waypoints import get_bundle_from_waypoints


def test_get_bundle_from_waypointsB(
    data_regression: DataRegressionFixture,
    check: bool = True,
) -> Component:

    ys1 = np.array([0, 5, 10, 15, 30, 40, 50, 60]) + 0.0
    ys2 = np.array([0, 10, 20, 30, 70, 90, 110, 120]) + 500.0
    N = ys1.size

    ports1 = [
        Port(name=f"A_{i}", midpoint=(0, ys1[i]), width=0.5, orientation=0)
        for i in range(N)
    ]
    ports2 = [
        Port(
            name=f"B_{i}",
            midpoint=(500, ys2[i]),
            width=0.5,
            orientation=180,
        )
        for i in range(N)
    ]

    p0 = ports1[0].position

    c = gf.Component("B")
    c.add_ports(ports1)
    c.add_ports(ports2)
    waypoints = [
        p0 + (200, 0),
        p0 + (200, -200),
        p0 + (400, -200),
        (p0[0] + 400, ports2[0].y),
    ]

    routes = get_bundle_from_waypoints(ports1, ports2, waypoints)
    lengths = {}
    for i, route in enumerate(routes):
        c.add(route.references)
        lengths[i] = route.length

    if check:
        data_regression.check(lengths)
    return c


def test_get_bundle_from_waypointsC(
    data_regression: DataRegressionFixture,
    check: bool = True,
) -> Component:

    ys1 = np.array([0, 5, 10, 15, 20, 60, 70, 80, 120, 125])
    ys2 = np.array([0, 5, 10, 20, 25, 30, 40, 55, 60, 65]) - 500.0
    N = ys1.size

    ports1 = [
        Port(name=f"A_{i}", midpoint=(0, ys1[i]), width=0.5, orientation=0)
        for i in range(N)
    ]
    ports2 = [
        Port(
            name=f"B_{i}",
            midpoint=(600, ys2[i]),
            width=0.5,
            orientation=180,
        )
        for i in range(N)
    ]

    c = gf.Component()
    c.add_ports(ports1)
    c.add_ports(ports2)
    waypoints = [
        ports1[0].position + (200, 0),
        ports1[0].position + (200, -200),
        ports1[0].position + (400, -200),
        (ports1[0].x + 400, ports2[0].y),
        ports2[0].position,
    ]

    routes = get_bundle_from_waypoints(ports1, ports2, waypoints)
    lengths = {}
    for i, route in enumerate(routes):
        c.add(route.references)
        lengths[i] = route.length

    if check:
        data_regression.check(lengths)

    return c


def test_get_bundle_from_waypoints_staggered(
    data_regression: DataRegressionFixture,
    check: bool = True,
):

    c = gf.Component()
    r = c << gf.components.array(
        component=gf.components.straight, rows=2, columns=1, spacing=(0, 20)
    )
    r.movex(60)
    r.movey(40)

    lt = c << gf.components.straight(length=15)
    lb = c << gf.components.straight(length=5)
    lt.movey(5)

    ports1 = lt.get_ports_list(orientation=0) + lb.get_ports_list(orientation=0)
    ports2 = r.get_ports_list(orientation=180)

    dx = 20
    p0 = ports1[0].midpoint + (dx, 0)
    p1 = (ports1[0].midpoint[0] + dx, ports2[0].midpoint[1])
    waypoints = (p0, p1)

    routes = gf.routing.get_bundle_from_waypoints(ports1, ports2, waypoints=waypoints)
    lengths = {}
    for i, route in enumerate(routes):
        c.add(route.references)
        lengths[i] = route.length

    if check:
        data_regression.check(lengths)

    return c


if __name__ == "__main__":
    # c = test_get_bundle_from_waypointsC(None, check=False)
    # c = test_get_bundle_from_waypointsB(None, check=False)
    c = test_get_bundle_from_waypoints_staggered(None, check=False)
    c.show()



"""Read Sweep parameters from YAML file and writes gdsfiles"""

import importlib
import io
import sys
from pathlib import Path
from typing import List, Union

from gdsfactory.config import CONFIG
from gdsfactory.sweep.read_sweep import read_sweep
from gdsfactory.sweep.write_sweep import write_sweep


def import_custom_doe_factories():
    """Find if we have custom DOEs on this config.
    Make them available in component_factory
    """

    sys.path += [CONFIG["mask_root_directory"]]
    if CONFIG["custom_components"]:
        try:
            importlib.import_module(CONFIG["custom_components"])
        except Exception:
            pass


def write_sweep_from_yaml(yaml: Union[str, Path]) -> List[List[Path]]:
    """Read DOE settings from yaml and writes GDS files build_directory

    Args:
        filepath: YAML string or filepath describing DOEs

    Returns:
        gdspaths: list

    For each DOE save:

    - GDS
    - json metadata
    - ports CSV
    - markdown report, with DOE settings
    """
    yaml = io.StringIO(yaml) if isinstance(yaml, str) and "\n" in yaml else yaml
    does = read_sweep(yaml)

    gdspaths = []
    for doe_name, doe in does.items():
        # print(doe_name)
        # print(sweep.get("settings"))
        # print(sweep.get("do_permutations"))
        # print(sweep)
        # print(list(sweep.keys()))
        # print(type(sweep.get('settings')))
        # assert type(sweep.get('settings'))
        d = write_sweep(
            component_type=doe.get("component"),
            doe_name=doe_name,
            do_permutations=doe.get("do_permutations", True),
            list_settings=doe.get("settings"),
            description=doe.get("description"),
            analysis=doe.get("analysis"),
            test=doe.get("test"),
            functions=doe.get("functions"),
        )
        gdspaths.append(d)
    return gdspaths


def test_write_doe_from_yaml() -> None:
    does_path = CONFIG["samples_path"] / "mask" / "does.yml"
    gdspaths = write_sweep_from_yaml(does_path)
    # print(len(gdspaths))
    assert len(gdspaths) == 4  # 2 does
    assert len(gdspaths[0]) == 2  # 2 GDS in the first DOE
    assert len(gdspaths[1]) == 2  # 2 GDS in the 2nd DOE


sample_yaml = """
mmi_width:
  component: mmi1x2
  settings:
    width_mmi: [4.5, 5.6]
    length_mmi: 10

mmi_width_length:
  component: mmi1x2
  do_permutation: True
  settings:
    length_mmi: [11, 12]
    width_mmi: [3.6, 7.8]

"""


def test_write_doe_from_yaml_string() -> None:
    gdspaths = write_sweep_from_yaml(sample_yaml)
    # print(len(gdspaths))
    assert len(gdspaths) == 2  # 2 does
    assert len(gdspaths[0]) == 2  # 2 GDS in the first DOE
    assert len(gdspaths[1]) == 4  # 4 GDS in the 2nd DOE


if __name__ == "__main__":
    test_write_doe_from_yaml()
    # test_write_doe_from_yaml_string()
    # gdspaths = write_doe_from_yaml(sample_yaml)
    # print(len(gdspaths))
    # print(len(gdspaths[0]))
    # print(len(gdspaths[1]))



from typing import Optional, Tuple

import picwriter.components as pc

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.waveguide_template import strip
from gdsfactory.types import ComponentFactory, Layer


@gf.cell
def spiral(
    port_spacing: float = 500.0,
    length: float = 10e3,
    spacing: Optional[float] = None,
    parity: int = 1,
    port: Tuple[int, int] = (0, 0),
    direction: str = "WEST",
    waveguide_template: ComponentFactory = strip,
    layer: Layer = gf.LAYER.WG,
    layer_cladding: Layer = gf.LAYER.WGCLAD,
    cladding_offset: float = 3.0,
    wg_width: float = 0.5,
    radius: float = 10.0,
    **kwargs
) -> Component:
    """Picwriter Spiral

    Args:
        port_spacing: distance between input/output ports
        length: spiral length (um)
        spacing: distance between parallel straights
        parity: If 1 spiral on right side, if -1 spiral on left side (mirror flip)
        port: Cartesian coordinate of the input port
        direction: NORTH, WEST, SOUTH, EAST  or angle in radians
        waveguide_template (WaveguideTemplate): Picwriter WaveguideTemplate function
        layer: core layer
        layer_cladding: cladding layer
        cladding_offset: distance from core to cladding
        wg_width: 0.5
        radius: 10
        kwargs: cross_section settings

    """
    c = pc.Spiral(
        gf.call_if_func(
            waveguide_template,
            wg_width=wg_width,
            radius=radius,
            layer=layer,
            layer_cladding=layer_cladding,
            cladding_offset=cladding_offset,
            **kwargs
        ),
        width=port_spacing,
        length=length,
        spacing=spacing,
        parity=parity,
        port=port,
        direction=direction,
    )
    return gf.read.from_picwriter(c, port_layer=layer)


if __name__ == "__main__":
    c = spiral(length=10e3, port_spacing=500, radius=20, direction="NORTH")
    c = spiral(length=10e3, port_spacing=500, radius=20, direction="EAST")
    c.show()



"""
The choice of font file is configurable from the YAML file
"""
import functools
from pathlib import Path
from typing import Dict, Tuple

import klayout.db as pya
from klayout.dbcore import Cell

import gdsfactory as gf
from gdsfactory.autoplacer.helpers import import_cell, load_gds

FONT_PATH = gf.CONFIG.get("font_path")


@functools.lru_cache()
def load_alphabet(filepath: Path = FONT_PATH) -> Dict[str, Cell]:
    c = load_gds(filepath)
    return {_c.name: _c for _c in c.each_cell()}


def add_text(
    cell: Cell,
    text: str,
    position: Tuple[int, int] = (0, 0),
    align_x: str = "center",
    align_y: str = "top",
    fontpath: Path = FONT_PATH,
) -> Cell:
    """add text label"""
    text = text.upper()
    alphabet = load_alphabet(filepath=fontpath)
    idbu = 1 / cell.layout().dbu
    si = alphabet["A"].dbbox()
    x, y = position
    w = si.width()
    h = si.height()
    c = cell.layout().create_cell("TEXT_{}".format(text))
    n = len(text)

    if align_x == "center":
        dx = -(n + 0.5) * w / 2
    elif align_x == "right":
        dx = -n * w
    else:
        dx = 0

    if align_y == "top":
        dy = -h
    elif align_y == "center":
        dy = -h / 2
    else:
        dy = 0

    for i, char in enumerate(text):
        _l = import_cell(cell.layout(), alphabet[char])
        _transform = pya.DTrans((i * w + dx) * idbu, dy * idbu)
        label = pya.CellInstArray(_l.cell_index(), _transform)
        c.insert(label)

    cell.insert(pya.CellInstArray(c.cell_index(), pya.DTrans(x, y)))
    return c


def test_alphabet() -> None:
    ly = pya.Layout()
    top_cell = ly.create_cell("TOP")
    add_text(top_cell, "HELLO-WORLD_0123456789+_-")
    # add_text(top_cell, "HELLO - WORLD 0123456789 +_-") # Need to add spaces
    # to the font
    top_cell.write("hello_world.gds")


if __name__ == "__main__":
    test_alphabet()



import gdsfactory as gf


@gf.cell
def rectangles(widths: gf.types.Floats) -> gf.Component:
    c = gf.Component()
    for width in widths:
        c << gf.components.rectangle(size=(width, width))

    c.distribute()
    return c


def test_name_partial_functions():
    s1 = gf.partial(gf.components.straight)
    s2 = gf.partial(gf.components.straight, length=5)
    s3 = gf.partial(gf.components.straight, 5)

    m1 = gf.partial(gf.components.mzi, straight=s1)()
    m2 = gf.partial(gf.components.mzi, straight=s2)()
    m3 = gf.partial(gf.components.mzi, straight=s3)()

    # print(m1.name)
    # print(m2.name)
    # print(m3.name)

    assert (
        m2.name == m3.name
    ), f"{m2.name} different from {m2.name} while they are the same function"
    assert (
        m1.name != m2.name
    ), f"{m1.name} is the same {m2.name} while they are different functions"
    assert (
        m1.name != m3.name
    ), f"{m1.name} is the same {m3.name} while they are different functions"


def test_name_iterators():
    c1 = rectangles([i for i in range(5)])
    c2 = rectangles([i for i in range(6)])
    assert c1.name != c2.name


def test_name_int_float():
    """Ensure that passing the same int or float returns the same component"""
    c1 = gf.components.straight(length=5)
    c2 = gf.components.straight(length=5.0)

    assert c1.name == c2.name, f"{c1.name} does not match {c2.name}"


def test_float_point_errors():
    c1 = gf.components.straight(length=5.0 + 1e-20)  # any unit below pm disappears
    c2 = gf.components.straight(length=5.0)
    assert c1.name == c2.name, f"{c1.name} does not match {c2.name}"


# def test_name_different_signatures():
#     c1 = gf.components.compass()

#     @gf.cell
#     def compass(layer=(2, 0)):
#         return gf.components.compass(layer=layer)

#     c2 = compass()
#     print(c1.name)
#     print(c2.name)

#     assert c1.name != c2.name, f"{c1.name} should differ from {c2.name}"


if __name__ == "__main__":
    test_name_partial_functions()
    # test_name_int_float()
    # test_name_different_signatures()

    # c1 = gf.components.compass()

    # @gf.cell
    # def compass(layer=(2, 0)):
    #     return gf.components.compass(layer=layer)

    # c2 = compass()
    # assert c1.name != c2.name, f"{c1.name} should differ from {c2.name}"



"""Tidy3D """

import concurrent.futures
import hashlib
from typing import Awaitable

import tidy3d as td
from tidy3d import web

import gdsfactory as gf
from gdsfactory.config import PATH, logger
from gdsfactory.types import PathType

_executor = concurrent.futures.ThreadPoolExecutor()


def get_sim_hash(sim: td.Simulation) -> str:
    """Returns simulation hash as the unique ID."""
    return hashlib.md5(str(sim).encode()).hexdigest()


def _get_results(
    sim: td.Simulation,
    dirpath: PathType = PATH.results_tidy3d,
    overwrite: bool = False,
) -> td.SimulationData:
    """Return SimulationData results from simulation.

    Only submits simulation if results not found locally or remotely.
    First tries to load simulation results from disk.
    Then it tries to load them from the server storage.
    Finally, submits simulation to run remotely

    Args:
        sim: Simulation
        dirpath: to store results locally
        overwrite: overwrites the data even if path exists
    """
    task_name = sim_hash = get_sim_hash(sim)
    sim_path = dirpath / f"{sim_hash}.hdf5"
    logger.info(f"running simulation {sim_hash}")

    hash_to_id = {d["task_name"][:32]: d["task_id"] for d in web.get_tasks()}
    filepath = dirpath / f"{sim_hash}.hdf5"
    job = web.Job(simulation=sim, task_name=task_name)

    # Results in local storage
    if sim_path.exists():
        logger.info(f"{sim_path!r} found in local storage")
        sim_data = td.SimulationData.from_file(filepath)

    # Results in server storage
    elif sim_hash in hash_to_id:
        task_id = hash_to_id[sim_hash]
        web.monitor(task_id)
        sim_data = web.load(
            task_id=task_id, path=str(filepath), replace_existing=overwrite
        )

    # Run simulation if results not in local or server storage
    else:
        sim_data = job.run(path=filepath)
    return sim_data


def get_results(
    sim: td.Simulation,
    dirpath=PATH.results_tidy3d,
    overwrite: bool = True,
) -> Awaitable[td.SimulationData]:
    """Return a SimulationData from Simulation.

    Works with Pool of threads.
    Each thread can run in paralell and only becomes blocking when you ask
    for the result

    Args:
        sim: Simulation
        dirpath: to store results locally
        overwrite: overwrites the data even if path exists


    .. code::
        import gdsfactory.simulation.tidy3d as gt

        component = gf.components.straight(length=3)
        sim = gt.get_simulation(component=component)
        sim_data = gt.get_results(sim).result()

    """
    return _executor.submit(_get_results, sim, dirpath, overwrite)


if __name__ == "__main__":
    import gdsfactory.simulation.gtidy3d as gt

    component = gf.components.straight(length=3)
    sim = gt.get_simulation(component=component)
    r = sim_data = get_results(sim=sim).result()
